# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""TCP transport for MeshCore companion devices over WiFi.

Uses the same frame format as USB serial.
"""

import contextlib
import logging
import socket

from gi.repository import GLib
from meshy.transport_base import (
    RX_MARKER,
    TX_MARKER,
    ConnectionState,
    FrameTransport,
)

log = logging.getLogger(__name__)

TCP_DEFAULT_PORT = 5000
TCP_CONNECT_TIMEOUT = 10  # seconds


class TcpDevice:
    """Represents a TCP endpoint for connection view compatibility."""

    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port
        self.path = f"{host}:{port}"
        self.address = self.path
        self.name = f"TCP {host}:{port}"

    def __repr__(self):
        return f"TcpDevice({self.path})"


class TcpTransport(FrameTransport):
    """TCP transport for MeshCore companion devices over WiFi."""

    _LABEL = "TCP"
    # Accept both markers — meshcore-proxy sends responses framed with 0x3C
    _RX_MARKERS = frozenset((RX_MARKER, TX_MARKER))

    def __init__(self):
        super().__init__()
        self._socket: socket.socket | None = None

    def get_paired_devices(self) -> list:
        """TCP doesn't have paired devices."""
        return []

    def connect_device(self, device_path: str):
        """Connect to a TCP endpoint. device_path is 'host:port'."""
        if self._state == ConnectionState.CONNECTED:
            self.disconnect()

        self._set_state(ConnectionState.CONNECTING)

        try:
            host, port_str = device_path.rsplit(":", 1)
            port = int(port_str)
        except (ValueError, AttributeError):
            log.error(f"Invalid TCP address: {device_path}")
            self._set_state(ConnectionState.DISCONNECTED)
            return

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(TCP_CONNECT_TIMEOUT)
            sock.connect((host, port))
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            sock.settimeout(None)  # blocking reads in thread
            self._socket = sock
            self._connected_address = device_path
            log.info(f"Connected to TCP {host}:{port}")

            self._start_read_thread()
            self._set_state(ConnectionState.CONNECTED)

        except Exception as e:
            log.error(f"TCP connect failed ({device_path}): {e}")
            if self._socket:
                with contextlib.suppress(Exception):
                    self._socket.close()
                self._socket = None
            self._set_state(ConnectionState.DISCONNECTED)

    def disconnect(self):
        """Disconnect from the TCP endpoint."""
        if self._state in (ConnectionState.DISCONNECTED, ConnectionState.DISCONNECTING):
            return

        self._set_state(ConnectionState.DISCONNECTING)
        self._running = False

        if self._socket:
            with contextlib.suppress(Exception):
                self._socket.shutdown(socket.SHUT_RDWR)
        self._close_io()
        self._connected_address = None
        self._set_state(ConnectionState.DISCONNECTED)

    # ─── FrameTransport implementation ────────────────────────

    def _is_io_ready(self) -> bool:
        return self._socket is not None

    def _write_bytes(self, frame: bytes):
        self._socket.sendall(frame)

    def _close_io(self):
        if self._socket:
            with contextlib.suppress(Exception):
                self._socket.close()
            self._socket = None

    def _read_loop(self, gen: int):
        """Background thread that reads from the TCP socket."""
        sock = self._socket
        while self._running and gen == self._connect_gen and sock:
            try:
                data = sock.recv(4096)
                if not data:
                    if self._running and gen == self._connect_gen:
                        log.info("TCP: remote closed connection")
                        GLib.idle_add(self._handle_disconnect, gen)
                    break
                self._rx_buffer.extend(data)
                self._process_rx_buffer()
            except OSError as e:
                if self._running and gen == self._connect_gen:
                    log.error(f"TCP read error: {e}")
                    GLib.idle_add(self._handle_disconnect, gen)
                break



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Gdk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Adw, Gdk, Gio, GLib, Gtk
from meshy.background_portal import BackgroundPortal

log = logging.getLogger(__name__)


class MeshyApplication(Adw.Application):
    def __init__(self, version="26.07", **kwargs):
        super().__init__(
            application_id="page.codeberg.sesivany.Meshy",
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
            **kwargs,
        )
        self.version = version
        self._window = None
        self._pending_uri = None  # URI received before window is ready
        self._hidden_start = False
        self._background_held = False
        self._bg_portal = BackgroundPortal()

        self.add_main_option(
            "debug",
            ord("d"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            "Enable debug logging",
            None,
        )
        self.add_main_option(
            "hidden",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            "Start without showing the window (used for autostart)",
            None,
        )

    def do_handle_local_options(self, options):
        if options.contains("debug"):
            logging.basicConfig(level=logging.DEBUG, format="%(name)s %(levelname)s %(message)s")
        else:
            logging.basicConfig(level=logging.WARNING)
            # On macOS, also log to file for troubleshooting (console output hidden when launched via Finder)
            from meshy.platforms import is_macos

            if is_macos():
                import os

                log_file = os.path.expanduser("~/Library/Logs/Meshy/debug.log")
                os.makedirs(os.path.dirname(log_file), exist_ok=True)
                file_handler = logging.FileHandler(log_file, mode="w")
                file_handler.setLevel(logging.DEBUG)
                file_handler.setFormatter(
                    logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")
                )
                logging.getLogger().addHandler(file_handler)
                logging.getLogger().setLevel(logging.DEBUG)
        self._hidden_start = options.contains("hidden")
        return -1  # continue normal startup

    def do_activate(self):
        if not self._window:
            from meshy.window import MeshyWindow

            self._window = MeshyWindow(application=self)
            settings = Gio.Settings(schema_id="page.codeberg.sesivany.Meshy")
            if settings.get_boolean("run-in-background"):
                self._enable_background_mode()
        if self._hidden_start:
            self._hidden_start = False
            return

        # Activate NSApplication on macOS before presenting window
        from meshy.platforms import is_macos

        if is_macos():
            from meshy.platforms.macos.app_activation import activate_macos_app

            activate_macos_app()

        self._window.present()
        # Process any URI that arrived before the window was created
        if self._pending_uri:
            uri = self._pending_uri
            self._pending_uri = None
            GLib.idle_add(lambda: self._handle_meshcore_uri(uri) or False)

    def do_open(self, files, n_files, hint):
        """Handle meshcore:// URIs opened from the system."""
        self.do_activate()
        for f in files:
            uri = f.get_uri()
            if uri and uri.startswith("meshcore://"):
                if self._window and self._window.is_connected:
                    self._handle_meshcore_uri(uri)
                else:
                    self._pending_uri = uri

    def _handle_meshcore_uri(self, uri):
        """Process a meshcore:// URI -- import contact or channel."""
        if not self._window:
            return
        import re
        from urllib.parse import parse_qs, urlparse

        type_labels = {1: _("Chat"), 2: _("Repeater"), 3: _("Room"), 4: _("Sensor")}

        # Raw hex format: meshcore://<hex> — no slashes, no query params
        hex_data = uri[len("meshcore://") :].strip("/")
        if re.fullmatch(r"[0-9a-fA-F]+", hex_data):
            try:
                contact_frame = bytes.fromhex(hex_data)
                name, contact_type = self._parse_advert_name(contact_frame)
                type_label = type_labels.get(contact_type, _("Unknown"))
                if name:
                    body = _("Name: {name}\nType: {type}").format(name=name, type=type_label)
                else:
                    body = _("Type: {type}\nKey: {key}…").format(
                        type=type_label, key=hex_data[4:36]
                    )
                self._show_import_dialog(
                    _("Import Contact"),
                    body,
                    lambda: (
                        self._window.import_contact(contact_frame),
                        self._window.show_toast(_("Contact {}").format(name or _("imported"))),
                    ),
                )
            except ValueError:
                self._window.show_toast(_("Invalid meshcore:// link"))
            return

        parsed = urlparse(uri)
        params = parse_qs(parsed.query)

        # meshcore://channel/add?name=...&secret=...
        full_path = f"{parsed.netloc}{parsed.path}".rstrip("/")
        if full_path == "channel/add" and "secret" in params:
            ch_name = params.get("name", [""])[0]
            secret_hex = params["secret"][0].strip()
            try:
                secret = bytes.fromhex(secret_hex)
            except ValueError:
                self._window.show_toast(_("Invalid channel secret in link"))
                return
            if len(secret) != 16:
                self._window.show_toast(_("Invalid channel secret length"))
                return
            self._show_import_dialog(
                _("Add Channel"),
                _("Name: {name}\nType: Private").format(name=ch_name or _("Unnamed")),
                lambda: self._window.add_channel_from_qr(ch_name, secret),
            )
            return

        # meshcore://contact/add?name=...&public_key=...&type=...
        if "public_key" in params:
            pub_key_hex = params["public_key"][0].strip()
            name = params.get("name", [""])[0]
            try:
                contact_type = int(params.get("type", ["1"])[0])
            except ValueError:
                contact_type = 1
            try:
                pub_key = bytes.fromhex(pub_key_hex)
            except ValueError:
                self._window.show_toast(_("Invalid public key in link"))
                return
            if len(pub_key) != 32:
                self._window.show_toast(_("Invalid public key length"))
                return
            if self._window.is_contact_known(pub_key_hex):
                self._window.show_toast(
                    _("{} is already in your list").format(name or _("Contact"))
                )
                return
            type_label = type_labels.get(contact_type, _("Chat"))
            self._show_import_dialog(
                _("Add {} Contact").format(type_label),
                _("Name: {name}\nPublic Key: {key}…").format(
                    name=name or _("Unknown"), key=pub_key_hex[:16]
                ),
                lambda: (
                    self._window.add_contact(pub_key, contact_type, name or "Unknown"),
                    self._window.show_toast(_("Contact {} added").format(name or pub_key_hex[:16])),
                ),
            )
            return

        self._window.show_toast(_("Unrecognized meshcore:// link"))

    @staticmethod
    def _parse_advert_name(frame: bytes) -> tuple:
        """Extract name and type from a raw advert packet frame.

        Returns (name, adv_type) or ('', 0) if parsing fails.
        Frame layout: header(1) + path_len(1) + path(var) + pubkey(32)
                      + timestamp(4) + signature(64) + appdata(flags + ...)
        """
        try:
            i = 0
            frame[i]
            i += 1
            path_len_byte = frame[i]
            i += 1
            hop_count = path_len_byte & 0x3F
            hash_size = ((path_len_byte >> 6) & 0x03) + 1
            i += hop_count * hash_size  # skip path
            i += 32  # pub key
            i += 4  # timestamp
            i += 64  # signature
            if i >= len(frame):
                return ("", 0)
            flags = frame[i]
            i += 1
            adv_type = flags & 0x0F
            if flags & 0x10:  # has latlon
                i += 8
            if flags & 0x20:  # has feat1
                i += 2
            if flags & 0x40:  # has feat2
                i += 2
            if flags & 0x80:  # has name
                name = frame[i:].decode("utf-8", errors="replace").rstrip("\x00")
                return (name, adv_type)
            return ("", adv_type)
        except (IndexError, ValueError):
            return ("", 0)

    def _show_import_dialog(self, heading, body, on_confirm):
        """Show a confirmation dialog before importing."""
        dialog = Adw.AlertDialog(heading=heading, body=body)
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("add", _("Add"))
        dialog.set_response_appearance("add", Adw.ResponseAppearance.SUGGESTED)
        dialog.connect("response", lambda d, r: on_confirm() if r == "add" else None)
        dialog.present(self._window)

    def do_startup(self):
        # Register GResource before chaining up so GTK can auto-discover
        # gtk/help-overlay.ui during Gtk.Application.do_startup().
        self._load_gresource()
        Adw.Application.do_startup(self)
        self._setup_icon_theme()
        self._load_css()
        self._setup_actions()
        self._apply_saved_theme()

    def _load_gresource(self):
        """Load the bundled GResource file."""
        import os

        import meshy as _meshy_pkg

        resource_path = os.path.join(_meshy_pkg.pkgdatadir, "meshy.gresource")
        if os.path.exists(resource_path):
            resource = Gio.Resource.load(resource_path)
            Gio.resources_register(resource)

    def _setup_icon_theme(self):
        """Add custom icon resource path (requires display, call after startup)."""
        icon_theme = Gtk.IconTheme.get_for_display(Gdk.Display.get_default())
        icon_theme.add_resource_path("/page/codeberg/sesivany/Meshy/icons")

    def _load_css(self):
        from meshy.platforms import is_macos

        # Load base CSS from resource
        provider = Gtk.CssProvider()
        provider.load_from_resource("/page/codeberg/sesivany/Meshy/style.css")
        Gtk.StyleContext.add_provider_for_display(
            Gdk.Display.get_default(),
            provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

        # Add macOS-specific CSS for traffic light buttons
        if is_macos():
            macos_css = b"""
        /* macOS traffic light button spacing */
        .macos-window .nav-rail {
            /* Add space at top for traffic lights (28px buttons + padding) */
            padding-top: 40px;
        }
        .macos-window #sidebar_header {
            /* Add left padding to avoid traffic light overlap */
            padding-left: 80px;
        }
        .macos-window #connection_header {
            /* Connection screen header also needs spacing */
            padding-left: 80px;
        }
        """
            macos_provider = Gtk.CssProvider()
            macos_provider.load_from_data(macos_css)
            Gtk.StyleContext.add_provider_for_display(
                Gdk.Display.get_default(),
                macos_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
            )

    def _request_background_portal(self):
        settings = Gio.Settings(schema_id="page.codeberg.sesivany.Meshy")
        autostart = settings.get_boolean("autostart")
        self._bg_portal.request_background(
            _("Receiving messages in the background"),
            autostart,
            ["meshy", "--hidden"] if autostart else [],
            self._on_background_response,
        )

    def _enable_background_mode(self):
        if not self._background_held:
            self.hold()
            self._background_held = True
        if self._window:
            self._window.set_background_mode(True)
        self._request_background_portal()

    def _disable_background_mode(self):
        if self._background_held:
            self.release()
            self._background_held = False
        if self._window:
            self._window.set_background_mode(False)
        self._bg_portal.request_background(
            _("Receiving messages in the background"),
            False,
            [],
            lambda bg, auto: None,
        )
        self._bg_portal.clear_status()

    def _on_background_response(self, bg_granted, autostart_granted):
        log.info("Background portal response: bg=%s autostart=%s", bg_granted, autostart_granted)
        if not bg_granted and self._background_held:
            self.release()
            self._background_held = False
            if self._window:
                self._window.set_background_mode(False)

    def update_background_mode(self, enabled):
        if enabled:
            self._enable_background_mode()
        else:
            self._disable_background_mode()

    def update_autostart(self, enabled):
        if not enabled:
            self._bg_portal.request_background(
                _("Receiving messages in the background"),
                False,
                [],
                self._on_background_response,
            )
        elif self._background_held:
            self._request_background_portal()

    def _do_full_quit(self, *_args):
        if self._window:
            self._window.shutdown()
        if self._background_held:
            self.release()
            self._background_held = False
        self._bg_portal.clear_status()
        self.quit()

    def _setup_actions(self):
        quit_action = Gio.SimpleAction.new("quit", None)
        quit_action.connect("activate", self._do_full_quit)
        self.add_action(quit_action)
        self.set_accels_for_action("win.quit", ["<primary>q"])

        about_action = Gio.SimpleAction.new("about", None)
        about_action.connect("activate", self._on_about)
        self.add_action(about_action)

        import meshy as _meshy_pkg

        if _meshy_pkg.SHORTCUTS_DIALOG_ENABLED:
            self.set_accels_for_action("win.show-help-overlay", ["<primary>question"])
        self.set_accels_for_action("win.disconnect", ["<primary>d"])
        self.set_accels_for_action("win.help", ["F1"])
        self.set_accels_for_action("win.navigate-device", ["<primary>1"])
        self.set_accels_for_action("win.navigate-contacts", ["<primary>2"])
        self.set_accels_for_action("win.navigate-channels", ["<primary>3"])
        self.set_accels_for_action("win.navigate-map", ["<primary>4"])
        self.set_accels_for_action("win.navigate-settings", ["<primary>5"])
        self.set_accels_for_action("win.search", ["<primary>f"])
        self.set_accels_for_action("win.new-message", ["<primary>n"])
        self.set_accels_for_action("win.navigate-prev", ["<Alt>Up"])
        self.set_accels_for_action("win.navigate-next", ["<Alt>Down"])
        self.set_accels_for_action("win.navigate-prev-unread", ["<Alt><Shift>Up"])
        self.set_accels_for_action("win.navigate-next-unread", ["<Alt><Shift>Down"])
        self.set_accels_for_action("win.custom-tiles", ["<primary>m"])

    def _apply_saved_theme(self):
        from meshy.theme_manager import ThemeManager

        ThemeManager.get_default().apply_from_settings()

    def _on_about(self, action, param):
        about = Adw.AboutDialog(
            application_name="Meshy",
            application_icon="page.codeberg.sesivany.Meshy",
            version=self.version,
            developer_name="Jiri Eischmann",
            developers=[
                "Jiri Eischmann",
                "Jaroslav Řezník",
                "stereo",
                "vanous",
                "Moritz Bitsch",
                "Petr Menšík",
                "Manuel Traut",
                "Lukáš Viktora",
                "clauwn",
            ],
            translator_credits=_(
                "gallegonovato — Spanish\n"
                "lebeno — French, Dutch\n"
                "bjawebos — German\n"
                "Manuel Traut — German\n"
                "lejun — French\n"
                "LievenBlancke — Dutch\n"
                "sesivany — Czech"
            ),
            license_type=Gtk.License.GPL_3_0,
            comments=_("A GTK client for MeshCore LoRa mesh network devices"),
            website="https://meshy-app.org/",
            support_url="https://meshy-app.org/docs/",
            issue_url="https://codeberg.org/sesivany/meshy/issues",
            copyright="\u00a9 2026 Jiri Eischmann",
        )
        about.add_legal_section(
            "libshumate",
            "\u00a9 libshumate contributors",
            Gtk.License.LGPL_2_1,
            None,
        )
        about.add_legal_section(
            "zbar",
            "\u00a9 zbar contributors",
            Gtk.License.LGPL_2_1,
            None,
        )
        about.add_legal_section(
            "protobuf-c",
            "\u00a9 protobuf-c contributors",
            Gtk.License.BSD,
            None,
        )
        about.add_legal_section(
            "PyCryptodome",
            "\u00a9 PyCryptodome contributors",
            Gtk.License.BSD,
            None,
        )
        about.add_legal_section(
            "pyzbar",
            "\u00a9 pyzbar contributors",
            Gtk.License.MIT_X11,
            None,
        )
        about.add_legal_section(
            "pySerial",
            "\u00a9 pySerial contributors",
            Gtk.License.BSD,
            None,
        )
        about.add_legal_section(
            "Segno",
            "\u00a9 Segno contributors",
            Gtk.License.BSD,
            None,
        )
        about.add_legal_section(
            _("Adwaita Symbolic Icons"),
            "\u00a9 GNOME Project",
            Gtk.License.LGPL_3_0,
            None,
        )
        about.present(self._window)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import os as _os

pkgdatadir = _os.path.join(_os.path.dirname(_os.path.dirname(__file__)), "share", "meshy")
QR_SCANNER_ENABLED = True
SHORTCUTS_DIALOG_ENABLED = True



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""BLE transport factory for cross-platform support.

This module provides a platform-agnostic interface for BLE communication.
The actual implementation is selected at runtime based on the platform:
- Linux: BlueZ D-Bus API
- macOS: CoreBluetooth framework (to be implemented)
"""

from meshy.platforms import is_linux, is_macos

# Re-export BleDevice for backwards compatibility
from meshy.platforms.ble_base import BleDevice

__all__ = ["BleDevice", "BleTransport", "create_ble_transport"]


def create_ble_transport():
    """Factory function to create platform-appropriate BLE transport.

    Returns:
        BleTransportBase: Platform-specific BLE transport implementation

    Note: On macOS, if CoreBluetooth fails to initialize (metaclass conflict),
    this returns None and BLE features will be unavailable.
    """
    if is_macos():
        try:
            from meshy.platforms.macos.ble_corebluetooth import BleTransportCoreBluetooth

            # Use PyObjC alloc/init pattern for NSObject subclasses
            return BleTransportCoreBluetooth.alloc().init()
        except (TypeError, ImportError) as e:
            import logging

            log = logging.getLogger(__name__)
            log.warning(f"BLE not available on macOS: {e}")
            log.info("USB and TCP connectivity still available")
            return None
    elif is_linux():
        from meshy.platforms.linux.ble_bluez import BleTransportBlueZ

        return BleTransportBlueZ()
    else:
        import logging

        logging.getLogger(__name__).info(
            "BLE not available on this platform, USB and TCP connectivity still available"
        )
        return None


class BleTransport:
    """BLE transport factory class for backwards compatibility.

    Usage:
        transport = BleTransport()  # Returns platform-specific implementation
    """

    def __new__(cls):
        return create_ble_transport()



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Backup and restore for MeshCore device configuration and messages."""

import uuid
from datetime import datetime

from gi.repository import GLib
from meshy.models import (
    ChannelMessage,
    Message,
    MessageStatus,
    NotificationLevel,
)
from meshy.protocol import (
    build_add_update_contact,
    build_device_query,
    build_set_auto_add_config,
    build_set_channel,
)


class BackupManager:
    """Stateless helper for building and applying backup dicts."""

    def export_backup(self, win) -> dict:
        si = win.self_info
        di = win.device_info

        backup = {
            "name": di.name or si.get("name", ""),
            "public_key": di.public_key or si.get("public_key", ""),
            "radio_settings": {
                "frequency": int(di.radio_freq * 1000)
                if di.radio_freq
                else si.get("radio_freq", 0) * 1000,
                "bandwidth": int(di.radio_bw * 1000)
                if di.radio_bw
                else si.get("radio_bw", 0) * 1000,
                "spreading_factor": di.radio_sf or si.get("radio_sf", 0),
                "coding_rate": di.radio_cr or si.get("radio_cr", 0),
                "tx_power": di.tx_power or si.get("tx_power", 0),
            },
            "position_settings": {
                "latitude": str(si.get("adv_lat", 0.0)),
                "longitude": str(si.get("adv_lon", 0.0)),
            },
            "other_settings": {
                "manual_add_contacts": 1 if si.get("manual_add_contacts") else 0,
                "advert_location_policy": si.get("adv_loc_policy", 0),
            },
            "auto_add_settings": win.get_auto_add_config(),
        }

        backup["channels"] = []
        for ch in win.channels:
            if not ch.is_empty:
                backup["channels"].append(
                    {
                        "name": ch.name,
                        "secret": ch.psk_hex,
                    }
                )

        backup["contacts"] = []
        for c in win.contacts:
            entry = {
                "type": c.type,
                "name": c.name,
                "public_key": c.public_key_hex,
                "flags": c.flags,
                "latitude": str(c.latitude) if c.latitude else "0.0",
                "longitude": str(c.longitude) if c.longitude else "0.0",
                "last_advert": int(c.last_seen.timestamp()) if c.last_seen else 0,
                "last_modified": int(c.device_lastmod.timestamp()) if c.device_lastmod else 0,
                "out_path": ",".join(c.path_hops) if c.path_hops else None,
            }
            backup["contacts"].append(entry)

        if win.storage:
            backup["messages"] = {}
            for c in win.contacts:
                msgs = win.storage.get_messages(c.public_key_hex, limit=10000)
                if msgs:
                    backup["messages"][c.public_key_hex] = [
                        {
                            "text": m.text,
                            "timestamp": m.timestamp.timestamp(),
                            "is_outgoing": m.is_outgoing,
                            "sender_key_hex": m.sender_key_hex,
                            "room_sender_name": m.room_sender_name,
                        }
                        for m in msgs
                    ]

            backup["channel_messages"] = {}
            for ch in win.channels:
                if ch.is_empty:
                    continue
                msgs = win.storage.get_channel_messages(ch.index, limit=10000)
                if msgs:
                    backup["channel_messages"][str(ch.index)] = [
                        {
                            "sender_name": m.sender_name,
                            "text": m.text,
                            "timestamp": m.timestamp.timestamp(),
                            "is_outgoing": m.is_outgoing,
                        }
                        for m in msgs
                    ]

            passwords = {}
            for c in win.contacts:
                pw = win.storage.get_room_password(c.public_key_hex)
                if pw:
                    passwords[c.public_key_hex] = pw
            if passwords:
                backup["room_passwords"] = passwords

            notif = {}
            for ch in win.channels:
                if not ch.is_empty and ch.notification_level != NotificationLevel.DEFAULT:
                    notif[str(ch.index)] = ch.notification_level.value
            if notif:
                backup["notification_levels"] = notif

        return backup

    def import_backup(self, win, backup: dict, restore_messages: bool = True):
        imported_contacts = 0
        imported_channels = 0

        settings_cmds = []
        name = backup.get("name", "")
        if name:
            settings_cmds.append(lambda: win.set_advert_name(name))

        radio = backup.get("radio_settings", {})
        if radio:
            freq = radio.get("frequency", 0)
            bw = radio.get("bandwidth", 0)
            sf = radio.get("spreading_factor", 0)
            cr = radio.get("coding_rate", 0)
            tx = radio.get("tx_power", 0)
            if freq and bw and sf and cr:
                settings_cmds.append(lambda: win.set_radio_params(freq, bw, sf, cr))
            if tx:
                settings_cmds.append(lambda: win.set_tx_power(tx))

        pos = backup.get("position_settings", {})
        if pos:
            try:
                lat = float(pos.get("latitude", 0))
                lon = float(pos.get("longitude", 0))
                if abs(lat) > 1e-6 or abs(lon) > 1e-6:
                    settings_cmds.append(lambda: win.set_advert_latlon(lat, lon))
            except (ValueError, TypeError):
                pass

        other = backup.get("other_settings", {})
        if other:
            loc_policy = other.get("advert_location_policy", None)
            if loc_policy is not None:
                settings_cmds.append(lambda: win.set_advert_loc_policy(loc_policy))

        auto = backup.get("auto_add_settings", {})
        if auto:
            frame = build_set_auto_add_config(
                chat=auto.get("auto_add_chat", False),
                repeater=auto.get("auto_add_repeater", False),
                room=auto.get("auto_add_room_server", False),
                sensor=auto.get("auto_add_sensor", False),
                overwrite_oldest=auto.get("overwrite_oldest", False),
                max_hops=auto.get("max_hops", 0),
            )
            settings_cmds.append(lambda: win.send_frame(frame))

        for i, cmd in enumerate(settings_cmds):
            GLib.timeout_add(i * 300, lambda c=cmd: c() or False)
        settings_delay = len(settings_cmds) * 300

        # Optimistically update device info and settings UI
        di = win._device_info
        if name:
            di.name = name
        if radio:
            if freq and bw and sf and cr:
                di.radio_freq = freq
                di.radio_bw = bw
                di.radio_sf = sf
                di.radio_cr = cr
            if tx:
                di.tx_power = tx
        if pos:
            try:
                lat_v = float(pos.get("latitude", 0))
                lon_v = float(pos.get("longitude", 0))
                if abs(lat_v) > 1e-6 or abs(lon_v) > 1e-6:
                    di.latitude = lat_v
                    di.longitude = lon_v
            except (ValueError, TypeError):
                pass
        win._settings_view.update_device_info(di)

        device_cmds = []
        channels = backup.get("channels", [])
        used = {c.index for c in win.channels if not c.is_empty}
        for ch_data in channels:
            ch_name = ch_data.get("name", "")
            secret_hex = ch_data.get("secret", "")
            if ch_name and secret_hex and len(secret_hex) == 32:
                try:
                    psk = bytes.fromhex(secret_hex)
                    already = any(c.psk_hex == secret_hex for c in win.channels if not c.is_empty)
                    if already:
                        continue
                    slot = None
                    for idx in range(win.device_info.max_channels or 8):
                        if idx not in used:
                            slot = idx
                            break
                    if slot is not None:
                        device_cmds.append(
                            lambda s=slot, n=ch_name, p=psk: win.send_frame(
                                build_set_channel(s, n, p)
                            )
                        )
                        imported_channels += 1
                        used.add(slot)
                except ValueError:
                    pass

        contacts = backup.get("contacts", [])
        contacts_by_key = {c.public_key_hex: c for c in win.contacts}
        for c_data in contacts:
            pub_key_hex = c_data.get("public_key", "")
            if not pub_key_hex or len(pub_key_hex) != 64:
                continue
            if pub_key_hex in contacts_by_key:
                continue
            try:
                pub_key = bytes.fromhex(pub_key_hex)
                c_name = c_data.get("name", "Unknown")
                contact_type = c_data.get("type", 1)
                flags = c_data.get("flags", 0)
                lat = float(c_data.get("latitude", 0))
                lon = float(c_data.get("longitude", 0))
                last_mod = c_data.get("last_modified")
                device_cmds.append(
                    lambda pk=pub_key,
                    ct=contact_type,
                    fl=flags,
                    n=c_name,
                    la=lat,
                    lo=lon,
                    lm=last_mod: win.send_frame(
                        build_add_update_contact(
                            pub_key=pk,
                            contact_type=ct,
                            flags=fl,
                            name=n,
                            lat=la if abs(la) > 1e-6 else None,
                            lon=lo if abs(lo) > 1e-6 else None,
                            last_modified=lm,
                        )
                    )
                )
                imported_contacts += 1
            except (ValueError, TypeError):
                pass

        for i, cmd in enumerate(device_cmds):
            GLib.timeout_add(settings_delay + i * 300, lambda c=cmd: c() or False)

        if restore_messages and win.storage:
            with win.storage.batch():
                for pub_key_hex, msgs in backup.get("messages", {}).items():
                    for m in msgs:
                        ts = m.get("timestamp", 0)
                        text = m.get("text", "")
                        if win.storage.message_exists(pub_key_hex, ts, text):
                            continue
                        message = Message(
                            sender_key=bytes.fromhex(m.get("sender_key_hex", pub_key_hex)),
                            text=text,
                            timestamp=datetime.fromtimestamp(ts),
                            is_outgoing=m.get("is_outgoing", False),
                            status=MessageStatus.DELIVERED,
                            message_id=str(uuid.uuid4()),
                            room_sender_name=m.get("room_sender_name", ""),
                        )
                        win.storage.save_message(pub_key_hex, message)

                for ch_idx_str, msgs in backup.get("channel_messages", {}).items():
                    ch_idx = int(ch_idx_str)
                    for m in msgs:
                        msg = ChannelMessage(
                            channel_index=ch_idx,
                            sender_name=m.get("sender_name", ""),
                            text=m.get("text", ""),
                            timestamp=datetime.fromtimestamp(m.get("timestamp", 0)),
                            is_outgoing=m.get("is_outgoing", False),
                        )
                        win.storage.save_channel_message(msg)

                for pub_key_hex, pw in backup.get("room_passwords", {}).items():
                    win.storage.save_room_password(pub_key_hex, pw)

        total_delay = settings_delay + len(device_cmds) * 300 + 1000
        GLib.timeout_add(total_delay, lambda: win._frame_handler.send_get_contacts() or False)
        GLib.timeout_add(total_delay + 500, lambda: win._fetch_channels_sequential(0) or False)
        GLib.timeout_add(total_delay + 1000, lambda: win.send_frame(build_device_query()) or False)

        return imported_contacts, imported_channels



from unittest.mock import patch

from meshy.models import Channel, Contact


def _insert_msg(storage, contact_hex, text, ts, is_outgoing=False):
    storage._db.execute(
        """INSERT INTO messages
        (contact_key_hex, sender_key_hex, text, timestamp, is_outgoing, is_cli, status, message_id, room_sender_name)
        VALUES (?, ?, ?, ?, ?, 0, 0, NULL, '')""",
        (contact_hex, contact_hex, text, ts, int(is_outgoing)),
    )
    storage._db.commit()


CONTACT_HEX = "aa" * 32
OTHER_HEX = "bb" * 32


# ─── has_duplicate_message ─────────────────────────────────────


class TestHasDuplicateMessage:
    def test_empty_db_returns_false(self, storage):
        assert storage.has_duplicate_message(CONTACT_HEX, 1000.0, "hello") is False

    def test_layer1_exact_timestamp_match(self, storage):
        _insert_msg(storage, CONTACT_HEX, "hello", 1000.0)
        assert storage.has_duplicate_message(CONTACT_HEX, 1000.0, "hello") is True

    def test_layer1_different_text_not_duplicate(self, storage):
        _insert_msg(storage, CONTACT_HEX, "hello", 1000.0)
        assert storage.has_duplicate_message(CONTACT_HEX, 1000.0, "world") is False

    @patch("meshy.storage.time")
    def test_layer2_manual_resend_within_30min(self, mock_time, storage):
        ts = 1000.0
        mock_time.time.return_value = ts + 600  # 10 min later
        _insert_msg(storage, CONTACT_HEX, "hello", ts)
        assert storage.has_duplicate_message(CONTACT_HEX, ts + 60, "hello") is True

    @patch("meshy.storage.time")
    def test_layer2_not_duplicate_after_30min(self, mock_time, storage):
        ts = 1000.0
        mock_time.time.return_value = ts + 2000  # > 30 min
        _insert_msg(storage, CONTACT_HEX, "hello", ts)
        assert storage.has_duplicate_message(CONTACT_HEX, ts + 60, "hello") is False

    @patch("meshy.storage.time")
    def test_layer2_not_duplicate_after_our_reply(self, mock_time, storage):
        ts = 1000.0
        mock_time.time.return_value = ts + 300
        _insert_msg(storage, CONTACT_HEX, "hello", ts)
        _insert_msg(storage, CONTACT_HEX, "reply", ts + 100, is_outgoing=True)
        assert storage.has_duplicate_message(CONTACT_HEX, ts + 200, "hello") is False

    @patch("meshy.storage.time")
    def test_layer2_different_text_not_duplicate(self, mock_time, storage):
        ts = 1000.0
        mock_time.time.return_value = ts + 300
        _insert_msg(storage, CONTACT_HEX, "hello", ts)
        assert storage.has_duplicate_message(CONTACT_HEX, ts + 60, "world") is False


# ─── message_exists ────────────────────────────────────────────


class TestMessageExists:
    def test_within_5s_window(self, storage):
        _insert_msg(storage, CONTACT_HEX, "test", 1000.0)
        assert storage.message_exists(CONTACT_HEX, 1003.0, "test") is True

    def test_outside_5s_window(self, storage):
        _insert_msg(storage, CONTACT_HEX, "test", 1000.0)
        assert storage.message_exists(CONTACT_HEX, 1010.0, "test") is False

    def test_different_text(self, storage):
        _insert_msg(storage, CONTACT_HEX, "test", 1000.0)
        assert storage.message_exists(CONTACT_HEX, 1000.0, "other") is False


# ─── count_unread_messages ─────────────────────────────────────


class TestCountUnread:
    def test_all_unread_without_last_read(self, storage):
        storage.save_contact(Contact(public_key=bytes.fromhex(CONTACT_HEX), name="Alice"))
        _insert_msg(storage, CONTACT_HEX, "m1", 100.0)
        _insert_msg(storage, CONTACT_HEX, "m2", 200.0)
        _insert_msg(storage, CONTACT_HEX, "m3", 300.0, is_outgoing=True)
        assert storage.count_unread_messages(CONTACT_HEX) == 2

    def test_partial_unread_with_last_read(self, storage):
        storage.save_contact(Contact(public_key=bytes.fromhex(CONTACT_HEX), name="Alice"))
        _insert_msg(storage, CONTACT_HEX, "m1", 100.0)
        _insert_msg(storage, CONTACT_HEX, "m2", 200.0)
        _insert_msg(storage, CONTACT_HEX, "m3", 300.0)
        storage.set_contact_last_read_at(CONTACT_HEX, 150.0)
        assert storage.count_unread_messages(CONTACT_HEX) == 2

    def test_channel_unread(self, storage):
        storage.save_channel(Channel(index=0, name="General", psk=bytes(16)))
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) VALUES (0, 'Bob', 'hi', 100.0, 0)",
        )
        storage._db.execute(
            "INSERT INTO channel_messages (channel_index, sender_name, text, timestamp, is_outgoing) VALUES (0, 'Bob', 'there', 200.0, 0)",
        )
        storage._db.commit()
        assert storage.count_unread_channel_messages(0) == 2
        storage.set_channel_last_read_at(0, 150.0)
        assert storage.count_unread_channel_messages(0) == 1


# ─── upsert_path_record / get_path_history ─────────────────────


class TestPathHistory:
    def test_insert_new_record(self, storage):
        storage.upsert_path_record(CONTACT_HEX, "aabb", 1, 2, success=True, trip_ms=500)
        history = storage.get_path_history(CONTACT_HEX)
        assert len(history) == 1
        rec = history[0]
        assert rec["path_bytes"] == "aabb"
        assert rec["success_count"] == 1
        assert rec["failure_count"] == 0
        assert rec["last_trip_ms"] == 500

    def test_update_increments_counts(self, storage):
        storage.upsert_path_record(CONTACT_HEX, "aabb", 1, 2, success=True, trip_ms=500)
        storage.upsert_path_record(CONTACT_HEX, "aabb", 1, 2, success=True, trip_ms=300)
        storage.upsert_path_record(CONTACT_HEX, "aabb", 1, 2, success=False)
        history = storage.get_path_history(CONTACT_HEX)
        assert len(history) == 1
        assert history[0]["success_count"] == 2
        assert history[0]["failure_count"] == 1

    def test_history_limit_100(self, storage):
        for i in range(110):
            storage.upsert_path_record(CONTACT_HEX, f"path{i:03d}", 1, 1, success=True)
        history = storage.get_path_history(CONTACT_HEX)
        assert len(history) == 100


# ─── Migration / schema ───────────────────────────────────────


class TestMigration:
    def test_tables_created(self, storage):
        tables = {
            row[0]
            for row in storage._db.execute(
                "SELECT name FROM sqlite_master WHERE type='table'"
            ).fetchall()
        }
        for expected in (
            "contacts",
            "messages",
            "channels",
            "channel_messages",
            "path_history",
            "room_passwords",
            "settings",
            "regions",
        ):
            assert expected in tables, f"Missing table: {expected}"

    def test_contacts_has_migrated_columns(self, storage):
        cols = {row[1] for row in storage._db.execute("PRAGMA table_info(contacts)").fetchall()}
        for col in (
            "last_read_at",
            "path_hash_size",
            "device_lastmod",
            "force_flood",
            "auto_reset_path",
            "smart_route_selection",
        ):
            assert col in cols, f"Missing column: {col}"



import sys
from pathlib import Path

# The meshy package lives at src/ (imported as "meshy.*" by the installed app).
# For tests, make "import meshy.storage" work by pointing at src's parent
# with src aliased as "meshy" via a symlink-free approach.
_src = Path(__file__).resolve().parent.parent / "src"
if str(_src.parent) not in sys.path:
    sys.path.insert(0, str(_src.parent))
# Register src/ as the "meshy" package
import types

if "meshy" not in sys.modules:
    pkg = types.ModuleType("meshy")
    pkg.__path__ = [str(_src)]
    pkg.__package__ = "meshy"
    sys.modules["meshy"] = pkg

import pytest
from meshy.storage import Storage


@pytest.fixture
def storage():
    s = Storage(db_path=":memory:")
    yield s
    s.close()



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Adw, Gdk, Gio, Gtk

THEMES = {
    "meshcore": {
        "name": "MeshCore Dark",
        "foreground": (1.0, 1.0, 1.0, 0.9),
        "background": (0.067, 0.094, 0.153),
        "accent": (0.231, 0.510, 0.965),
        "is_dark": True,
    },
    "meshcore-light": {
        "name": "MeshCore Light",
        "foreground": (0.067, 0.094, 0.153, 0.85),
        "background": (0.945, 0.961, 0.976),
        "accent": (0.231, 0.510, 0.965),
        "is_dark": False,
    },
    "spring": {
        "name": "Spring",
        "foreground": (0.250, 0.368, 0.501),
        "background": (0.972, 1.0, 0.949),
        "accent": (0.2, 0.65, 0.4),
        "is_dark": False,
    },
    "midnight": {
        "name": "Midnight",
        "foreground": (1.0, 1.0, 1.0, 0.6),
        "background": (0.137, 0.207, 0.250),
        "accent": (0.160, 0.470, 0.650),
        "is_dark": True,
    },
    "gruvbox": {
        "name": "Gruvbox",
        "foreground": (0.984, 0.945, 0.780, 0.8),
        "background": (0.156, 0.156, 0.156),
        "accent": (0.407, 0.615, 0.415),
        "is_dark": True,
    },
    "ocean": {
        "name": "Ocean",
        "foreground": (0.741, 0.901, 0.984),
        "background": (0.117, 0.145, 0.160),
        "accent": (0.3, 0.7, 0.8),
        "is_dark": True,
    },
}

COLOR_SCHEME_MAP = {
    "system": Adw.ColorScheme.DEFAULT,
    "light": Adw.ColorScheme.FORCE_LIGHT,
    "dark": Adw.ColorScheme.FORCE_DARK,
}


class ThemeManager:
    _instance = None

    def __init__(self):
        self._recoloring_provider = Gtk.CssProvider()
        Gtk.StyleContext.add_provider_for_display(
            Gdk.Display.get_default(),
            self._recoloring_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION + 1,
        )
        self._settings = Gio.Settings(schema_id="page.codeberg.sesivany.Meshy")

    @classmethod
    def get_default(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def apply_from_settings(self):
        mode = self._settings.get_string("color-scheme")
        if mode == "custom":
            theme_id = self._settings.get_string("custom-theme")
            self.apply_custom_theme(theme_id)
        else:
            self.apply_color_scheme(mode)

    def apply_color_scheme(self, mode):
        scheme = COLOR_SCHEME_MAP.get(mode, Adw.ColorScheme.DEFAULT)
        Adw.StyleManager.get_default().set_color_scheme(scheme)
        self._recoloring_provider.load_from_string("")

    def apply_custom_theme(self, theme_id):
        theme = THEMES.get(theme_id)
        if not theme:
            Adw.StyleManager.get_default().set_color_scheme(Adw.ColorScheme.DEFAULT)
            self._recoloring_provider.load_from_string("")
            return
        if theme["is_dark"]:
            Adw.StyleManager.get_default().set_color_scheme(Adw.ColorScheme.FORCE_DARK)
        else:
            Adw.StyleManager.get_default().set_color_scheme(Adw.ColorScheme.FORCE_LIGHT)
        css = self._generate_css(theme)
        self._recoloring_provider.load_from_string(css)

    def set_color_scheme(self, mode):
        self._settings.set_string("color-scheme", mode)
        if mode != "custom":
            self.apply_color_scheme(mode)

    def set_custom_theme(self, theme_id):
        self._settings.set_string("custom-theme", theme_id)
        self._settings.set_string("color-scheme", "custom")
        self.apply_custom_theme(theme_id)

    @staticmethod
    def _rgba_str(components):
        if len(components) == 4:
            r, g, b, a = components
            return f"rgba({int(r*255)},{int(g*255)},{int(b*255)},{a:.2f})"
        r, g, b = components[:3]
        return f"rgb({int(r*255)},{int(g*255)},{int(b*255)})"

    @staticmethod
    def _rgba_opaque_str(components):
        r, g, b = components[:3]
        return f"rgb({int(r*255)},{int(g*255)},{int(b*255)})"

    def _generate_css(self, theme):
        fg = theme["foreground"]
        bg = theme["background"]
        accent = theme["accent"]
        is_dark = theme["is_dark"]

        alt = "white" if is_dark else "black"
        mix_level = 0.07 if is_dark else 0.25

        fg_str = self._rgba_str(fg)
        bg_str = self._rgba_opaque_str(bg)

        lines = []
        lines.append(f"@define-color window_bg_color {bg_str};")
        lines.append(f"@define-color window_fg_color {fg_str};")
        lines.append(f"@define-color headerbar_bg_color mix({bg_str},white,{mix_level});")
        lines.append(f"@define-color headerbar_fg_color {fg_str};")
        lines.append(f"@define-color popover_bg_color mix({bg_str},white,{mix_level});")
        lines.append(f"@define-color popover_fg_color {fg_str};")
        lines.append(
            f'@define-color view_bg_color mix({bg_str},{"black" if is_dark else "white"},{0.1 if is_dark else 0.3});'
        )
        lines.append(f"@define-color view_fg_color {fg_str};")
        lines.append(f"@define-color card_fg_color {fg_str};")
        lines.append(f"@define-color headerbar_border_color {fg_str};")
        lines.append(f"@define-color dialog_fg_color {fg_str};")
        lines.append("@define-color dark_fill_bg_color @headerbar_bg_color;")

        if is_dark:
            lines.append(f"@define-color dialog_bg_color mix({bg_str},white,0.07);")
            lines.append("@define-color card_bg_color alpha(white,0.08);")
        else:
            lines.append(f"@define-color dialog_bg_color {bg_str};")
            lines.append("@define-color card_bg_color alpha(white,0.8);")

        lines.append(f"@define-color sidebar_bg_color mix({bg_str},{alt},0.05);")
        lines.append(f"@define-color sidebar_backdrop_color mix({bg_str},{alt},0.1);")
        lines.append(f"@define-color sidebar_fg_color mix({fg_str},{alt},0.05);")
        lines.append(f"@define-color sidebar_shade_color mix({bg_str},black,0.1);")
        lines.append(f"@define-color sidebar_border_color mix({bg_str},black,0.1);")

        lines.append(f"@define-color secondary_sidebar_bg_color mix({bg_str},{alt},0.02);")
        lines.append(f"@define-color secondary_sidebar_backdrop_color mix({bg_str},{alt},0.05);")
        lines.append(f"@define-color secondary_sidebar_fg_color mix({fg_str},{alt},0.02);")
        lines.append(f"@define-color secondary_sidebar_shade_color mix({bg_str},black,0.1);")
        lines.append(f"@define-color secondary_sidebar_border_color mix({bg_str},black,0.1);")

        if accent:
            accent_str = self._rgba_opaque_str(accent)
            lines.append(f"@define-color accent_color {accent_str};")
            lines.append(f"@define-color accent_bg_color {accent_str};")
            lines.append("@define-color accent_fg_color white;")

        return "\n".join(lines)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Channels list and chat view."""

import gettext
import hashlib

ngettext = gettext.ngettext

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Shumate", "1.0")
from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gtk, Pango, Shumate
from meshy.models import Channel, ChannelMessage, NotificationLevel
from meshy.protocol import max_channel_text_bytes
from meshy.utils import extract_emoji, linkify, parse_shared_contact, parse_shared_location
from meshy.views import _add_hover_swap, _format_date_text
from meshy.views.chat_helpers import (
    CONTACT_TYPE_ICONS,
    ensure_loc_map,
    open_location_external,
    share_contact_dialog,
    share_location,
    share_location_map,
    show_location_map,
)
from meshy.views.chat_mixin import ChatScrollMixin

_BATCH_SIZE = 30


class ChannelChatItem(GObject.Object):
    """Model object for one row in the channel chat ListView."""

    KIND_MESSAGE = 0
    KIND_UNREAD_DIVIDER = 1
    KIND_DATE_DIVIDER = 2

    def __init__(self, kind=0, message=None, date_text=""):
        super().__init__()
        self.kind = kind
        self.message = message
        self.date_text = date_text
        self._repeat_count_val = 0
        self._resend_count_val = 0
        self._max_hops_val = -1

    @GObject.Property(type=int)
    def repeat_count(self):
        return self._repeat_count_val

    @repeat_count.setter
    def repeat_count(self, value):
        self._repeat_count_val = value

    @GObject.Property(type=int)
    def resend_count(self):
        return self._resend_count_val

    @resend_count.setter
    def resend_count(self, value):
        self._resend_count_val = value

    @GObject.Property(type=int)
    def max_hops(self):
        return self._max_hops_val

    @max_hops.setter
    def max_hops(self, value):
        self._max_hops_val = value


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/channel-chat-widget.ui")
class ChannelChatWidget(ChatScrollMixin, Gtk.Box):
    """Channel chat content widget - shown in the content panel."""

    __gtype_name__ = "MeshyChannelChatWidget"

    _stack = Gtk.Template.Child("stack")
    _scroll = Gtk.Template.Child("chat_scroll")
    _list_view = Gtk.Template.Child("list_view")
    _scroll_down_revealer = Gtk.Template.Child("scroll_down_revealer")
    _unread_btn_revealer = Gtk.Template.Child("unread_btn_revealer")
    _char_label = Gtk.Template.Child("char_label")
    _emoji_btn = Gtk.Template.Child("emoji_btn")
    _text_entry = Gtk.Template.Child("text_entry")
    _send_btn = Gtk.Template.Child("send_btn")

    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self._window = window
        self._current_channel: Channel | None = None
        # Track outgoing items for repeat detection: (channel_idx, timestamp, text) -> item
        self._outgoing_items: dict[tuple, ChannelChatItem] = {}
        # Repeat info per outgoing message key: list of {path_hops, path_len}
        self._repeats: dict[tuple, list] = {}
        # Track incoming message path variants: msg_key -> list of path_info dicts
        self._incoming_paths: dict[tuple, list] = {}
        self._seen_channel_msgs: set[tuple] = set()
        self._incoming_items: dict[tuple, ChannelChatItem] = {}
        # Map resend timestamps to original outgoing message keys
        # so echoes of resent messages can be matched back
        self._resend_ts_map: dict[tuple, tuple] = {}
        self._mention_popover = None
        self._mention_listbox = None

        # Pango attributes for char counter
        attrs = Pango.AttrList()
        attrs.insert(Pango.attr_scale_new(0.85))
        self._char_label.set_attributes(attrs)

        # ListView model and factory
        self._store = Gio.ListStore(item_type=ChannelChatItem)
        selection = Gtk.NoSelection(model=self._store)
        self._factory = Gtk.SignalListItemFactory()
        self._factory.connect("setup", self._on_factory_setup)
        self._factory.connect("bind", self._on_factory_bind)
        self._factory.connect("unbind", self._on_factory_unbind)
        self._list_view.set_model(selection)
        self._list_view.set_factory(self._factory)

        # Context menus (single popover on the ListView, menu switches per item)
        self._context_popover = Gtk.PopoverMenu()
        self._context_popover.set_parent(self._list_view)
        self._context_popover.set_has_arrow(False)
        self._context_popover.connect("closed", self._on_context_popover_closed)
        self._context_menu_open = False
        self._frozen_scroll_value = 0.0
        self._context_item = None

        self._outgoing_menu = Gio.Menu()
        self._outgoing_menu.append(_("Heard Repeats"), "msg.repeats")
        self._outgoing_menu.append(_("Send Again"), "msg.resend")
        self._outgoing_menu.append(_("Delete"), "msg.delete")

        self._incoming_menu = Gio.Menu()
        self._incoming_menu.append(_("Reply"), "msg.reply")
        self._incoming_menu.append(_("Copy Text"), "msg.copy")
        self._incoming_menu.append(_("View Message Paths"), "msg.paths")
        self._incoming_menu.append(_("Delete"), "msg.delete")

        ctx_group = Gio.SimpleActionGroup()
        for name, cb in [
            ("repeats", lambda a, p: self._show_repeats_context()),
            ("resend", lambda a, p: self._resend_context()),
            ("delete", lambda a, p: self._delete_context()),
            ("reply", lambda a, p: self._reply_context()),
            ("copy", lambda a, p: self._copy_context()),
            ("paths", lambda a, p: self._show_paths_context()),
        ]:
            action = Gio.SimpleAction.new(name, None)
            action.connect("activate", cb)
            ctx_group.add_action(action)
        self._list_view.insert_action_group("msg", ctx_group)

        # Gesture controllers
        right_click = Gtk.GestureClick(button=3)
        right_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        right_click.connect("pressed", self._on_list_right_click)
        self._list_view.add_controller(right_click)
        long_press = Gtk.GestureLongPress()
        long_press.set_touch_only(True)
        long_press.set_exclusive(True)
        long_press.connect("pressed", self._on_list_long_press)
        self._list_view.add_controller(long_press)

        self._is_sticky = True
        self._scroll_positions: dict[int, float] = {}
        self._restoring_scroll = False
        self._unread_pos = -1
        self._messages_offset = 0
        self._all_loaded = False
        self._loading_more = False

        # Signal connections
        self._scroll_down_revealer.get_child().connect("clicked", self._on_scroll_to_bottom_clicked)
        self._unread_btn_revealer.get_child().connect("clicked", self._on_scroll_to_unread_clicked)
        self._emoji_btn.connect("clicked", self._on_emoji_clicked)
        self._text_entry.connect("activate", self._on_send)
        self._text_entry.connect("changed", self._on_text_changed)
        mention_key_ctrl = Gtk.EventControllerKey()
        mention_key_ctrl.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        mention_key_ctrl.connect("key-pressed", self._on_entry_key_pressed)
        self._text_entry.add_controller(mention_key_ctrl)
        self._send_btn.connect("clicked", self._on_send)

        # Extra context menu on text entry
        extra_menu = Gio.Menu()
        extra_menu.append(_("Share Contact"), "entry.share-contact")
        extra_menu.append(_("Share Location"), "entry.share-location")
        extra_menu.append(_("Share Location from Map"), "entry.share-location-map")
        self._text_entry.set_extra_menu(extra_menu)
        entry_group = Gio.SimpleActionGroup()
        share_a = Gio.SimpleAction.new("share-contact", None)
        share_a.connect("activate", self._on_share_contact)
        entry_group.add_action(share_a)
        loc_a = Gio.SimpleAction.new("share-location", None)
        loc_a.connect("activate", self._on_share_location)
        entry_group.add_action(loc_a)
        loc_map_a = Gio.SimpleAction.new("share-location-map", None)
        loc_map_a.connect("activate", self._on_share_location_map)
        entry_group.add_action(loc_map_a)
        self._text_entry.insert_action_group("entry", entry_group)

        # Scroll state tracking
        adj = self._scroll.get_vadjustment()
        adj.connect("value-changed", self._on_scroll_value_changed)
        adj.connect("notify::upper", self._on_scroll_upper_changed)
        adj.connect("notify::page-size", self._on_scroll_upper_changed)
        self._scroll.connect("edge-overshot", self._on_edge_overshot)

        self._stack.set_visible_child_name("placeholder")

    # ─── Factory callbacks ───────────────────────────────────

    def _on_factory_setup(self, factory, list_item):
        """Create the widget tree for one list item (recycled)."""
        list_item.set_activatable(False)
        list_item.set_selectable(False)
        list_item.set_focusable(False)
        outer = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # ── Avatar ──
        avatar = Adw.Avatar(size=32, show_initials=True)
        avatar.set_valign(Gtk.Align.START)

        emoji_lbl = Gtk.Label()
        emoji_lbl.add_css_class("emoji-avatar")
        emoji_lbl.set_valign(Gtk.Align.START)

        avatar_stack = Gtk.Stack()
        avatar_stack.set_transition_type(Gtk.StackTransitionType.NONE)
        avatar_stack.add_named(avatar, "initials")
        avatar_stack.add_named(emoji_lbl, "emoji")
        avatar_stack.set_visible(False)

        # ── Message bubble ──
        msg_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=2,
            margin_end=12,
            margin_top=2,
            margin_bottom=2,
        )

        content_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL,
            spacing=8,
            margin_start=12,
        )
        content_box.append(avatar_stack)
        content_box.append(msg_box)

        # Sender name (incoming only)
        sender_lbl = Gtk.Label(xalign=0)
        sender_lbl.add_css_class("dim-label")
        sender_lbl.add_css_class("caption")
        sender_lbl.set_visible(False)
        msg_box.append(sender_lbl)

        # Card frame with text
        text_lbl = Gtk.Label(
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD_CHAR,
            xalign=0,
            max_width_chars=40,
        )
        text_lbl.set_margin_start(12)
        text_lbl.set_margin_end(12)
        text_lbl.set_margin_top(8)
        text_lbl.set_margin_bottom(8)

        frame = Gtk.Frame()
        frame.add_css_class("card")

        content_stack = Gtk.Stack()
        content_stack.set_transition_type(Gtk.StackTransitionType.NONE)
        content_stack.set_vhomogeneous(False)
        content_stack.set_hhomogeneous(False)
        content_stack.add_named(text_lbl, "text")

        sc_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=6,
            margin_start=12,
            margin_end=12,
            margin_top=8,
            margin_bottom=8,
        )
        sc_header_lbl = Gtk.Label(xalign=0, wrap=True)
        sc_header_lbl.add_css_class("dim-label")
        sc_header_lbl.add_css_class("caption")
        sc_box.append(sc_header_lbl)

        sc_name_box = Gtk.Box(spacing=6)
        sc_icon = Gtk.Image.new_from_icon_name("avatar-default-symbolic")
        sc_name_lbl = Gtk.Label(xalign=0)
        attrs = Pango.AttrList()
        attrs.insert(Pango.attr_weight_new(Pango.Weight.BOLD))
        sc_name_lbl.set_attributes(attrs)
        sc_name_box.append(sc_icon)
        sc_name_box.append(sc_name_lbl)
        sc_box.append(sc_name_box)

        sc_key_lbl = Gtk.Label(xalign=0)
        sc_key_lbl.add_css_class("dim-label")
        sc_key_lbl.add_css_class("caption")
        sc_box.append(sc_key_lbl)

        sc_add_btn = Gtk.Button(label=_("Add Contact"))
        sc_add_btn.add_css_class("suggested-action")
        sc_add_btn.set_halign(Gtk.Align.START)
        sc_add_btn.set_margin_top(4)
        sc_box.append(sc_add_btn)

        content_stack.add_named(sc_box, "contact")

        # ── Shared location card ──
        loc_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=6,
            margin_start=12,
            margin_end=12,
            margin_top=8,
            margin_bottom=8,
        )
        loc_header_lbl = Gtk.Label(xalign=0, wrap=True)
        loc_header_lbl.add_css_class("dim-label")
        loc_header_lbl.add_css_class("caption")
        loc_box.append(loc_header_lbl)

        loc_coords_lbl = Gtk.Label(xalign=0, selectable=True)
        loc_coords_lbl.add_css_class("caption")
        loc_box.append(loc_coords_lbl)

        loc_map_frame = Gtk.Frame()
        loc_map_frame.add_css_class("card")
        loc_map_frame.set_overflow(Gtk.Overflow.HIDDEN)
        loc_map_frame.set_cursor(Gdk.Cursor.new_from_name("pointer"))
        loc_map_gesture = Gtk.GestureClick()
        loc_map_frame.add_controller(loc_map_gesture)
        loc_box.append(loc_map_frame)

        loc_open_btn = Gtk.Button(label=_("Open in Maps App"), margin_top=4)
        loc_box.append(loc_open_btn)

        content_stack.add_named(loc_box, "location")

        frame.set_child(content_stack)
        msg_box.append(frame)

        # Meta line: time + resend + repeat
        meta_box = Gtk.Box(spacing=4)

        time_lbl = Gtk.Label()
        time_lbl.add_css_class("dim-label")
        time_lbl.add_css_class("caption")
        meta_box.append(time_lbl)

        resend_lbl = Gtk.Label(label="")
        resend_lbl.add_css_class("dim-label")
        resend_lbl.add_css_class("caption")
        resend_lbl.set_visible(False)
        meta_box.append(resend_lbl)

        repeat_lbl = Gtk.Label(label="")
        repeat_lbl.add_css_class("dim-label")
        repeat_lbl.add_css_class("caption")
        repeat_lbl.set_visible(False)
        meta_box.append(repeat_lbl)

        msg_box.append(meta_box)
        outer.append(content_box)

        # ── Divider ──
        div_box = Gtk.Box(
            spacing=8,
            margin_start=12,
            margin_end=12,
            margin_top=8,
            margin_bottom=8,
        )
        line_left = Gtk.Separator(hexpand=True, valign=Gtk.Align.CENTER)
        div_box.append(line_left)
        div_lbl = Gtk.Label()
        div_box.append(div_lbl)
        line_right = Gtk.Separator(hexpand=True, valign=Gtk.Align.CENTER)
        div_box.append(line_right)
        div_box.set_visible(False)
        outer.append(div_box)

        # Store widget references
        outer._content_box = content_box
        outer._avatar = avatar
        outer._emoji_lbl = emoji_lbl
        outer._avatar_stack = avatar_stack
        outer._msg_box = msg_box
        outer._sender_lbl = sender_lbl
        outer._text_lbl = text_lbl
        outer._frame = frame
        outer._time_lbl = time_lbl
        outer._resend_lbl = resend_lbl
        outer._repeat_lbl = repeat_lbl
        outer._meta_box = meta_box
        outer._div_box = div_box
        outer._div_lbl = div_lbl
        outer._div_line_left = line_left
        outer._div_line_right = line_right
        outer._content_stack = content_stack
        outer._sc_header_lbl = sc_header_lbl
        outer._sc_icon = sc_icon
        outer._sc_name_lbl = sc_name_lbl
        outer._sc_key_lbl = sc_key_lbl
        outer._sc_add_btn = sc_add_btn
        outer._sc_add_handler_id = None
        outer._loc_header_lbl = loc_header_lbl
        outer._loc_coords_lbl = loc_coords_lbl
        outer._loc_map_frame = loc_map_frame
        outer._loc_open_btn = loc_open_btn
        outer._loc_map_gesture = loc_map_gesture
        outer._loc_map_click_handler = None
        outer._loc_open_handler = None
        outer._loc_map_widget = None
        outer._loc_viewport = None
        outer._loc_marker = None
        outer._chat_item = None
        outer._handler_ids = []

        # Wrap in per-item Clamp for width limiting
        clamp = Adw.Clamp(maximum_size=600, child=outer)
        clamp._outer = outer
        list_item.set_child(clamp)

    def _on_factory_bind(self, factory, list_item):
        """Populate widgets from the model item."""
        item = list_item.get_item()
        clamp = list_item.get_child()
        outer = clamp._outer
        outer._chat_item = item

        if item.kind == ChannelChatItem.KIND_MESSAGE:
            msg = item.message
            outer._msg_box.set_visible(True)
            outer._div_box.set_visible(False)

            # Alignment
            if msg.is_outgoing:
                outer._content_box.set_halign(Gtk.Align.END)
                outer._meta_box.set_halign(Gtk.Align.END)
                outer._avatar_stack.set_visible(False)
            else:
                outer._content_box.set_halign(Gtk.Align.START)
                outer._meta_box.set_halign(Gtk.Align.START)
                # Avatar
                emoji = extract_emoji(msg.sender_name) if msg.sender_name else None
                if emoji:
                    outer._emoji_lbl.set_label(emoji)
                    outer._avatar_stack.set_visible_child_name("emoji")
                else:
                    outer._avatar.set_text(msg.sender_name or "")
                    outer._avatar_stack.set_visible_child_name("initials")
                outer._avatar_stack.set_visible(True)

            # Sender name
            if not msg.is_outgoing and msg.sender_name:
                outer._sender_lbl.set_label(msg.sender_name)
                outer._sender_lbl.set_visible(True)
            else:
                outer._sender_lbl.set_visible(False)

            # Text / shared contact
            sc = parse_shared_contact(msg.text)
            if sc:
                pub_key_hex, contact_type, contact_name = sc
                outer._content_stack.set_visible_child_name("contact")
                if msg.is_outgoing:
                    outer._sc_header_lbl.set_label(_("You shared a contact:"))
                else:
                    sender = msg.sender_name or "?"
                    outer._sc_header_lbl.set_label(
                        _("{name} shared a contact with you:").format(name=sender)
                    )
                outer._sc_icon.set_from_icon_name(
                    CONTACT_TYPE_ICONS.get(contact_type, "avatar-default-symbolic")
                )
                outer._sc_name_lbl.set_label(contact_name)
                outer._sc_key_lbl.set_label(f"{pub_key_hex[:8]}...{pub_key_hex[-8:]}")
                if msg.is_outgoing:
                    outer._sc_add_btn.set_visible(False)
                else:
                    outer._sc_add_btn.set_visible(True)
                    is_known = self._window.is_contact_known(pub_key_hex)
                    outer._sc_add_btn.set_sensitive(not is_known)
                    outer._sc_add_btn.set_label(
                        _("Already added") if is_known else _("Add Contact")
                    )
                    if outer._sc_add_handler_id is not None:
                        outer._sc_add_btn.disconnect(outer._sc_add_handler_id)
                    outer._sc_add_handler_id = outer._sc_add_btn.connect(
                        "clicked",
                        self._on_shared_contact_add,
                        pub_key_hex,
                        contact_type,
                        contact_name,
                    )
            else:
                loc = parse_shared_location(msg.text)
                if loc:
                    lat, lon = loc
                    outer._content_stack.set_visible_child_name("location")
                    if msg.is_outgoing:
                        outer._loc_header_lbl.set_label(_("You shared a location:"))
                    else:
                        sender = msg.sender_name or "?"
                        outer._loc_header_lbl.set_label(
                            _("{name} shared a location:").format(name=sender)
                        )
                    outer._loc_coords_lbl.set_label(f"{lat:.6f}, {lon:.6f}")
                    ensure_loc_map(outer, lat, lon)
                    outer._loc_map_click_handler = outer._loc_map_gesture.connect(
                        "released",
                        lambda g, n, x, y, la=lat, lo=lon: show_location_map(self._window, la, lo),
                    )
                    outer._loc_open_handler = outer._loc_open_btn.connect(
                        "clicked", open_location_external, lat, lon
                    )
                else:
                    outer._content_stack.set_visible_child_name("text")
                    outer._text_lbl.set_markup(linkify(msg.text))

            # Frame style
            if msg.is_outgoing:
                outer._frame.add_css_class("outgoing-message")
            else:
                outer._frame.remove_css_class("outgoing-message")

            # Timestamp
            outer._time_lbl.set_label(msg.timestamp.strftime("%H:%M"))

            # Outgoing: resend + repeat labels
            if msg.is_outgoing:
                outer._resend_lbl.set_visible(True)
                outer._repeat_lbl.set_visible(True)
                self._update_resend_label_widget(outer, item)
                self._update_repeat_label_widget(outer, item)
                h1 = item.connect(
                    "notify::repeat-count",
                    lambda obj, pspec, w=outer: self._update_repeat_label_widget(w, obj),
                )
                h2 = item.connect(
                    "notify::resend-count",
                    lambda obj, pspec, w=outer: self._update_resend_label_widget(w, obj),
                )
                outer._handler_ids = [h1, h2]
            else:
                outer._resend_lbl.set_visible(False)
                outer._repeat_lbl.set_visible(True)
                self._update_hops_label_widget(outer, item)
                h1 = item.connect(
                    "notify::max-hops",
                    lambda obj, pspec, w=outer: self._update_hops_label_widget(w, obj),
                )
                outer._handler_ids = [h1]

        elif item.kind == ChannelChatItem.KIND_DATE_DIVIDER:
            outer._msg_box.set_visible(False)
            outer._div_box.set_visible(True)
            outer._div_lbl.set_label(item.date_text)
            outer._div_lbl.remove_css_class("unread-divider")
            outer._div_lbl.add_css_class("dim-label")
            outer._div_line_left.remove_css_class("unread-divider-line")
            outer._div_line_left.add_css_class("date-divider-line")
            outer._div_line_right.remove_css_class("unread-divider-line")
            outer._div_line_right.add_css_class("date-divider-line")

        elif item.kind == ChannelChatItem.KIND_UNREAD_DIVIDER:
            outer._msg_box.set_visible(False)
            outer._div_box.set_visible(True)
            outer._div_lbl.set_label(_("New Messages"))
            outer._div_lbl.remove_css_class("dim-label")
            outer._div_lbl.add_css_class("unread-divider")
            outer._div_line_left.remove_css_class("date-divider-line")
            outer._div_line_left.add_css_class("unread-divider-line")
            outer._div_line_right.remove_css_class("date-divider-line")
            outer._div_line_right.add_css_class("unread-divider-line")

    def _on_factory_unbind(self, factory, list_item):
        """Clean up signal connections when the widget is recycled."""
        clamp = list_item.get_child()
        outer = clamp._outer
        for hid in outer._handler_ids:
            if outer._chat_item:
                outer._chat_item.disconnect(hid)
        outer._handler_ids = []
        if outer._sc_add_handler_id is not None:
            outer._sc_add_btn.disconnect(outer._sc_add_handler_id)
            outer._sc_add_handler_id = None
        if outer._loc_map_click_handler is not None:
            outer._loc_map_gesture.disconnect(outer._loc_map_click_handler)
            outer._loc_map_click_handler = None
        if outer._loc_open_handler is not None:
            outer._loc_open_btn.disconnect(outer._loc_open_handler)
            outer._loc_open_handler = None
        outer._chat_item = None

    # ─── Shared contact helpers ─────────────────────────────

    def _on_shared_contact_add(self, btn, pub_key_hex, contact_type, name):
        self._window.add_contact(bytes.fromhex(pub_key_hex), contact_type, name)
        btn.set_sensitive(False)
        btn.set_label(_("Already added"))

    # ─── Label update helpers ────────────────────────────────

    @staticmethod
    def _update_repeat_label_widget(outer, item):
        count = item.repeat_count
        if count > 0:
            outer._repeat_lbl.set_label(
                ngettext("\u00b7 %d repeat", "\u00b7 %d repeats", count) % count
            )
        else:
            outer._repeat_lbl.set_label("")

    @staticmethod
    def _update_hops_label_widget(outer, item):
        hops = item.max_hops
        if hops == 0:
            outer._repeat_lbl.set_label("\u00b7 " + _("direct"))
        elif hops == 255 or hops == -1:
            outer._repeat_lbl.set_label("")
        elif hops > 0:
            outer._repeat_lbl.set_label("\u00b7 " + ngettext("%d hop", "%d hops", hops) % hops)
        else:
            outer._repeat_lbl.set_label("")

    @staticmethod
    def _update_resend_label_widget(outer, item):
        count = item.resend_count
        if count > 0:
            outer._resend_lbl.set_label(f'\u00b7 {count} resend{"s" if count != 1 else ""}')
        else:
            outer._resend_lbl.set_label("")

    # ─── Context menu ───────────────────────────────────────

    def _on_list_right_click(self, gesture, n_press, x, y):
        outer = self._find_outer_at(x, y)
        if not outer:
            return
        item = outer._chat_item
        if item and item.kind == ChannelChatItem.KIND_MESSAGE and item.message:
            if gesture:
                gesture.set_state(Gtk.EventSequenceState.CLAIMED)
            self._context_item = item
            self._context_menu_open = True
            self._is_sticky = False
            self._frozen_scroll_value = self._scroll.get_vadjustment().get_value()
            if item.message.is_outgoing:
                self._context_popover.set_menu_model(self._outgoing_menu)
            else:
                self._context_popover.set_menu_model(self._incoming_menu)
            rect = Gdk.Rectangle()
            rect.x, rect.y, rect.width, rect.height = int(x), int(y), 1, 1
            self._context_popover.set_pointing_to(rect)
            self._context_popover.popup()

    def _show_repeats_context(self):
        item = self._context_item
        if not item or not item.message:
            return
        msg = item.message
        key = self._make_msg_key(msg.channel_index, msg.timestamp, msg.text)
        self._show_repeats_dialog(key)

    def _resend_context(self):
        item = self._context_item
        if not item or not item.message:
            return
        self._resend_message(item.message, item)

    def _delete_context(self):
        item = self._context_item
        if not item or not item.message:
            return
        self._delete_message_item(item)

    def _reply_context(self):
        item = self._context_item
        if not item or not item.message:
            return
        self._reply_to_message(item.message)

    def _copy_context(self):
        item = self._context_item
        if not item or not item.message:
            return
        self._copy_message_text(item.message)

    def _show_paths_context(self):
        item = self._context_item
        if not item or not item.message:
            return
        msg = item.message
        key = self._make_msg_key(msg.channel_index, msg.timestamp, msg.text)
        self._show_message_paths(key, msg)

    # ─── Emoji ───────────────────────────────────────────────

    def _on_share_contact(self, action, param):
        if not self._current_channel:
            return
        channel_idx = self._current_channel.index

        def _send(c):
            text = f"<{c.public_key_hex}:{c.type}:{c.name}>"
            self._window.send_channel_message(channel_idx, text)

        share_contact_dialog(self._window, self._window.contacts, _send)

    def _on_share_location(self, action, param):
        if not self._current_channel:
            return
        channel_idx = self._current_channel.index
        share_location(
            self._window, lambda text: self._window.send_channel_message(channel_idx, text)
        )

    def _on_share_location_map(self, action, param):
        if not self._current_channel:
            return
        channel_idx = self._current_channel.index
        share_location_map(
            self._window, lambda text: self._window.send_channel_message(channel_idx, text)
        )

    # ─── @Mention autocomplete ─────────────────────────────────

    def _get_channel_senders(self) -> list[str]:
        my_name = self._window.device_info.name or ""
        seen = set()
        result = []
        for i in range(self._store.get_n_items() - 1, -1, -1):
            item = self._store.get_item(i)
            if item.kind == ChannelChatItem.KIND_MESSAGE and item.message:
                name = item.message.sender_name
                if name and name != my_name and name not in seen:
                    seen.add(name)
                    result.append(name)
        return result

    def _update_mention_popover(self, prefix: str, at_char_idx: int):
        senders = self._get_channel_senders()
        filtered = (
            [n for n in senders if n.lower().startswith(prefix.lower())] if prefix else senders
        )

        if not filtered:
            if self._mention_popover:
                self._mention_popover.popdown()
            return

        if not self._mention_popover:
            self._mention_popover = Gtk.Popover()
            self._mention_popover.set_parent(self._text_entry)
            self._mention_popover.set_autohide(False)
            self._mention_popover.set_has_arrow(False)
            self._mention_popover.set_position(Gtk.PositionType.TOP)
            self._mention_popover.set_halign(Gtk.Align.START)
            self._mention_listbox = Gtk.ListBox()
            self._mention_listbox.set_selection_mode(Gtk.SelectionMode.SINGLE)
            self._mention_listbox.connect("row-activated", self._on_mention_selected)
            scrolled = Gtk.ScrolledWindow(
                propagate_natural_height=True,
                max_content_height=280,
                hscrollbar_policy=Gtk.PolicyType.NEVER,
            )
            scrolled.set_child(self._mention_listbox)
            self._mention_popover.set_child(scrolled)

        try:
            delegate = self._text_entry.get_delegate()
            layout = delegate.get_layout()
            index_bytes = len(self._text_entry.get_text()[:at_char_idx].encode("utf-8"))
            strong_pos, _ = layout.get_cursor_pos(index_bytes)
            layout_x, _ = delegate.get_layout_offsets()
            cursor_x = layout_x + strong_pos.x // Pango.SCALE
        except Exception:
            cursor_x = 0

        rect = Gdk.Rectangle()
        rect.x = cursor_x
        rect.y = 0
        rect.width = 1
        rect.height = self._text_entry.get_height()
        self._mention_popover.set_pointing_to(rect)

        while row := self._mention_listbox.get_first_child():
            self._mention_listbox.remove(row)
        for name in filtered[:20]:
            label = Gtk.Label(
                label=name, xalign=0, margin_start=8, margin_end=8, margin_top=4, margin_bottom=4
            )
            self._mention_listbox.append(label)

        self._mention_popover.popup()

    def _check_mention_trigger(self, text: str, cursor_pos: int):
        text_before = text[:cursor_pos]
        at_idx = text_before.rfind("@")
        if at_idx == -1:
            if self._mention_popover:
                self._mention_popover.popdown()
            return
        if at_idx > 0 and text_before[at_idx - 1] not in (" ", "\t"):
            if self._mention_popover:
                self._mention_popover.popdown()
            return
        after_at = text_before[at_idx + 1 :]
        if "[" in after_at:
            if self._mention_popover:
                self._mention_popover.popdown()
            return
        self._update_mention_popover(after_at, at_idx)

    def _on_mention_selected(self, listbox, row):
        label = row.get_child()
        name = label.get_label()
        text = self._text_entry.get_text()
        cursor_pos = self._text_entry.get_position()
        text_before = text[:cursor_pos]
        at_idx = text_before.rfind("@")
        if at_idx == -1:
            return
        mention = f"@[{name}] "
        new_text = text[:at_idx] + mention + text[cursor_pos:]
        self._text_entry.set_text(new_text)
        self._text_entry.set_position(at_idx + len(mention))
        self._mention_popover.popdown()

    def _on_entry_key_pressed(self, ctrl, keyval, keycode, state):
        if not self._mention_popover or not self._mention_popover.get_visible():
            return False
        if keyval == Gdk.KEY_Escape:
            self._mention_popover.popdown()
            return True
        if keyval in (Gdk.KEY_Down, Gdk.KEY_Up):
            selected = self._mention_listbox.get_selected_row()
            if keyval == Gdk.KEY_Down:
                if selected:
                    next_row = selected.get_next_sibling()
                    if next_row:
                        self._mention_listbox.select_row(next_row)
                else:
                    first = self._mention_listbox.get_first_child()
                    if first:
                        self._mention_listbox.select_row(first)
            elif keyval == Gdk.KEY_Up and selected:
                prev_row = selected.get_prev_sibling()
                if prev_row:
                    self._mention_listbox.select_row(prev_row)
            return True
        if keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter, Gdk.KEY_Tab):
            selected = self._mention_listbox.get_selected_row()
            if selected:
                self._on_mention_selected(self._mention_listbox, selected)
                return True
        return False

    # ─── Channel & messages ──────────────────────────────────

    def get_last_message_time(self, channel_index: int) -> float | None:
        try:
            if not self._window.storage:
                return None
            return self._window.storage.get_max_channel_message_timestamp(channel_index)
        except Exception:
            return None

    def set_channel(self, channel: Channel):
        if self._mention_popover:
            self._mention_popover.popdown()
        if self._current_channel is not None:
            adj = self._scroll.get_vadjustment()
            self._scroll_positions[self._current_channel.index] = adj.get_value()
        self._current_channel = channel
        self._stack.set_visible_child_name("chat")
        self._load_messages(channel)

    def _load_messages(self, channel: Channel):
        self._outgoing_items.clear()
        self._repeats.clear()
        self._incoming_paths.clear()
        self._seen_channel_msgs.clear()
        self._incoming_items.clear()
        self._resend_ts_map.clear()
        self._unread_pos = -1
        self._messages_offset = 0
        self._all_loaded = False
        self._loading_more = False

        messages = self._window.get_channel_messages(channel.index, limit=_BATCH_SIZE)
        self._messages_offset = len(messages)
        self._all_loaded = len(messages) < _BATCH_SIZE
        last_read_at = self._window.get_channel_last_read_at(channel.index)

        self._preload_paths(messages)

        new_items = []
        divider_inserted = False
        last_date = None
        for msg in messages:
            msg_date = msg.timestamp.date()
            if last_date is not None and msg_date != last_date:
                new_items.append(
                    ChannelChatItem(
                        kind=ChannelChatItem.KIND_DATE_DIVIDER,
                        date_text=_format_date_text(msg.timestamp),
                    )
                )
            last_date = msg_date
            if (
                not divider_inserted
                and last_read_at is not None
                and not msg.is_outgoing
                and msg.timestamp.timestamp() > last_read_at
            ):
                new_items.append(ChannelChatItem(kind=ChannelChatItem.KIND_UNREAD_DIVIDER))
                self._unread_pos = len(new_items) - 1
                divider_inserted = True
            new_items.append(self._create_message_item(msg))

        saved_pos = self._scroll_positions.get(channel.index)
        if saved_pos is not None:
            self._is_sticky = False
            self._restoring_scroll = True

        # Replace all items in a single atomic operation
        self._store.splice(0, self._store.get_n_items(), new_items)

        if saved_pos is not None:
            GLib.idle_add(lambda pos=saved_pos: self._restore_scroll_position(pos))
        else:
            GLib.idle_add(self._scroll_to_bottom)
        if self._unread_pos >= 0:
            GLib.idle_add(lambda: self._update_unread_btn_visibility() or False)

        # Mark as read now
        self._window.mark_channel_read(channel.index)

    def _preload_paths(self, messages):
        if not self._window.storage:
            return
        for msg in messages:
            key = self._make_msg_key(msg.channel_index, msg.timestamp, msg.text)
            if msg.is_outgoing:
                paths = self._window.storage.get_channel_message_paths(
                    msg.channel_index,
                    int(msg.timestamp.timestamp()),
                    msg.text,
                )
                if paths:
                    self._repeats[key] = [
                        {
                            "path_hops": self._split_path_hops(
                                p.get("path_bytes", ""),
                                p.get("path_len", 0),
                            ),
                            "path_len": p.get("path_len", 0),
                        }
                        for p in paths
                    ]
            else:
                if key not in self._incoming_paths:
                    paths = self._window.storage.get_channel_message_paths(
                        msg.channel_index,
                        int(msg.timestamp.timestamp()),
                        msg.text,
                    )
                    if paths:
                        self._incoming_paths[key] = paths

    def _load_older_messages(self):
        if self._all_loaded or self._loading_more or not self._current_channel:
            return
        self._loading_more = True

        messages = self._window.get_channel_messages(
            self._current_channel.index, limit=_BATCH_SIZE, offset=self._messages_offset
        )
        if not messages:
            self._all_loaded = True
            self._loading_more = False
            return

        self._preload_paths(messages)

        older_items = []
        last_date = None
        for msg in messages:
            msg_date = msg.timestamp.date()
            if last_date is not None and msg_date != last_date:
                older_items.append(
                    ChannelChatItem(
                        kind=ChannelChatItem.KIND_DATE_DIVIDER,
                        date_text=_format_date_text(msg.timestamp),
                    )
                )
            last_date = msg_date
            older_items.append(self._create_message_item(msg))

        # Add boundary date divider if the dates differ
        if last_date is not None:
            for i in range(self._store.get_n_items()):
                first = self._store.get_item(i)
                if first.kind == ChannelChatItem.KIND_MESSAGE:
                    if first.message.timestamp.date() != last_date:
                        older_items.append(
                            ChannelChatItem(
                                kind=ChannelChatItem.KIND_DATE_DIVIDER,
                                date_text=_format_date_text(first.message.timestamp),
                            )
                        )
                    break

        self._store.splice(0, 0, older_items)
        if self._unread_pos >= 0:
            self._unread_pos += len(older_items)
        self._messages_offset += len(messages)
        self._all_loaded = len(messages) < _BATCH_SIZE
        self._loading_more = False

    @staticmethod
    def _split_path_hops(path_bytes_hex: str, hop_count: int) -> list[str]:
        """Split concatenated path hex into individual hop prefixes.

        Supports 1-byte (2 hex), 2-byte (4 hex), and 3-byte (6 hex) hash sizes.
        The hash size is derived from total bytes / hop count.
        """
        if not path_bytes_hex or hop_count <= 0:
            return []
        total_bytes = len(path_bytes_hex) // 2
        if total_bytes % hop_count == 0:
            hash_size = total_bytes // hop_count
        else:
            hash_size = 1  # fallback
        chars_per_hop = hash_size * 2
        return [
            path_bytes_hex[i : i + chars_per_hop].upper()
            for i in range(0, len(path_bytes_hex), chars_per_hop)
        ]

    def _make_msg_key(self, channel_index, timestamp, text):
        """Create a key for matching outgoing messages to repeats."""
        ts = int(timestamp.timestamp()) if hasattr(timestamp, "timestamp") else int(timestamp)
        return (channel_index, ts, text)

    def _create_message_item(self, msg: ChannelMessage) -> ChannelChatItem:
        """Create a ChannelChatItem and register it in tracking dicts."""
        item = ChannelChatItem(kind=ChannelChatItem.KIND_MESSAGE, message=msg)
        key = self._make_msg_key(msg.channel_index, msg.timestamp, msg.text)

        if msg.is_outgoing:
            self._outgoing_items[key] = item
            if key not in self._repeats:
                self._repeats[key] = []
            # Set initial repeat count from persisted data
            item._repeat_count_val = len(self._repeats.get(key, []))
        else:
            self._incoming_items[key] = item
            paths = self._incoming_paths.get(key, [])
            if paths:
                item._max_hops_val = max(p.get("path_len", -1) for p in paths)

        return item

    def _add_message_item(self, msg: ChannelMessage):
        """Create a ChannelChatItem and add it to the store."""
        item = self._create_message_item(msg)
        self._store.append(item)

    # ─── Dialogs (unchanged) ─────────────────────────────────

    def _show_message_paths(self, key, msg):
        """Show dialog with message path variants and repeater hops."""
        paths = self._incoming_paths.get(key, [])
        # Fallback to storage if no in-memory data
        if not paths and self._window.storage:
            paths = self._window.storage.get_channel_message_paths(
                msg.channel_index,
                int(msg.timestamp.timestamp()),
                msg.text,
            )
            if paths:
                self._incoming_paths[key] = paths

        if not paths:
            dialog = Adw.AlertDialog(
                heading=_("Message Paths"),
                body=_(
                    "No path information available. Path data is only "
                    "captured for messages received while connected."
                ),
            )
            dialog.add_response("ok", _("OK"))
            dialog.present(self._window)
            return

        dialog = Adw.Dialog()
        dialog.set_title(_("Message Paths"))
        dialog.set_content_width(420)
        dialog.set_content_height(480)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        scroll = Gtk.ScrolledWindow(vexpand=True)
        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=8,
            margin_start=12,
            margin_end=12,
            margin_top=12,
            margin_bottom=12,
        )

        info_label = Gtk.Label(
            label=f'Message from {msg.sender_name or "unknown"} '
            f'received via {len(paths)} path{"s" if len(paths) != 1 else ""}:',
            wrap=True,
            xalign=0,
        )
        info_label.add_css_class("dim-label")
        content.append(info_label)

        for i, p in enumerate(paths, 1):
            snr = p.get("snr")
            path_len = p.get("path_len", -1)
            path_bytes_hex = p.get("path_bytes", "")

            if path_bytes_hex and path_len > 0:
                hops = self._split_path_hops(path_bytes_hex, path_len)
                num_hops = len(hops) if hops else path_len
                hops_str = f'{num_hops} hop{"s" if num_hops != 1 else ""}'
            elif path_len == 255 or path_len == -1:
                hops_str = "flood"
            elif path_len == 0:
                hops_str = "direct"
            else:
                hops_str = f'{path_len} hop{"s" if path_len != 1 else ""}'

            snr_str = f"SNR: {snr:.1f} dB" if snr is not None else ""
            title_parts = [f"Path {i}: {hops_str}"]
            if snr_str:
                title_parts.append(snr_str)

            path_group = Adw.PreferencesGroup(
                title=" \u00b7 ".join(title_parts),
            )

            if path_bytes_hex and path_len > 0:
                hops = self._split_path_hops(path_bytes_hex, path_len)
                hop_contacts = self._window.resolve_hop_prefixes(hops)
                for hop_idx, prefix in enumerate(hops):
                    contact = hop_contacts[hop_idx]
                    name = contact.name if contact else _("Unknown repeater")
                    subtitle = f"Hop {hop_idx + 1} \u00b7 prefix {prefix}"

                    hop_row = Adw.ActionRow(
                        title=name,
                        subtitle=subtitle,
                    )
                    hop_row.add_prefix(Gtk.Image.new_from_icon_name("network-server-symbolic"))
                    path_group.add(hop_row)

                if any(c and c.has_location for c in hop_contacts):
                    map_btn = Gtk.Button(icon_name="map-symbolic", valign=Gtk.Align.CENTER)
                    map_btn.add_css_class("flat")
                    map_btn.set_tooltip_text(_("Show on Map"))
                    map_btn.connect(
                        "clicked",
                        lambda b, hc=hop_contacts, hp=hops: self._show_channel_path_on_map(hc, hp),
                    )
                    path_group.set_header_suffix(map_btn)
            elif path_len == 0:
                direct_row = Adw.ActionRow(
                    title=_("Direct message"),
                    subtitle=_("No repeaters were used"),
                    sensitive=False,
                )
                direct_row.add_prefix(Gtk.Image.new_from_icon_name("network-wireless-symbolic"))
                path_group.add(direct_row)
            else:
                no_path_row = Adw.ActionRow(
                    title=_("No repeater details"),
                    subtitle=_("Path data is only captured via LOG_DATA while connected"),
                    sensitive=False,
                )
                path_group.add(no_path_row)

            content.append(path_group)

        scroll.set_child(content)
        toolbar_view.set_content(scroll)
        dialog.set_child(toolbar_view)
        dialog.present(self._window)

    def _show_channel_path_on_map(self, hop_contacts, hop_prefixes=None):
        """Show a channel message path on the map."""
        import gi
        from meshy.views.map_view import (
            _MARKER_COLORS,
            _MARKER_ICONS,
            _SELF_COLOR,
            _UNKNOWN_COLOR,
            MapView,
        )

        gi.require_version("Shumate", "1.0")

        device = self._window.device_info
        self_lat = getattr(device, "latitude", None)
        self_lon = getattr(device, "longitude", None)
        has_self = self_lat and self_lon and (abs(self_lat) > 1e-6 or abs(self_lon) > 1e-6)

        # Build full hop list including unknowns
        raw_hops = []
        for i, c in enumerate(hop_contacts):
            prefix = hop_prefixes[i] if hop_prefixes and i < len(hop_prefixes) else None
            if c and c.has_location:
                raw_hops.append((c.latitude, c.longitude, c, prefix))
            else:
                raw_hops.append((None, None, c, prefix))

        if has_self:
            raw_hops.append((self_lat, self_lon, None, None))

        # Interpolate unknown positions
        interp = []
        for lat, lon, contact, prefix in raw_hops:
            interp.append((lat, lon, contact, prefix, None))
        interp = MapView.interpolate_unknown_hops(interp)

        if not any(h[0] is not None for h in interp):
            return

        dialog, map_widget, marker_layer, viewport = MapView.open_map_dialog(
            self._window, title=_("Message Path")
        )

        map_widget.get_map().remove_layer(marker_layer)
        path_layer = MapView.create_dashed_path_layer(viewport)
        map_widget.get_map().add_layer(path_layer)
        map_widget.get_map().add_layer(marker_layer)

        lats, lons = [], []

        hop_num = 0
        for lat, lon, contact, prefix, _snr in interp:
            if lat is None:
                continue
            node = Shumate.Marker()
            node.set_location(lat, lon)
            path_layer.add_node(node)

            if contact is None and prefix is None:
                # Self/device marker
                r, g, b = _SELF_COLOR
                icon_name = "network-wireless-symbolic"
                name = device.name or _("My Device")
                type_label = None
                badge = ""
            elif contact is None or not contact.has_location:
                # Unknown/interpolated hop
                r, g, b = _UNKNOWN_COLOR
                icon_name = "network-server-symbolic"
                name = prefix or _("Unknown")
                type_label = _("Unknown repeater")
                hop_num += 1
                badge = str(hop_num)
            else:
                r, g, b = _MARKER_COLORS.get(contact.type, (0.5, 0.5, 0.5))
                icon_name = _MARKER_ICONS.get(contact.type, "avatar-default-symbolic")
                name = contact.name
                type_label = contact.type_label
                hop_num += 1
                badge = str(hop_num)

            is_interpolated = (contact is None or not contact.has_location) and prefix is not None
            detail_cb = (lambda c=contact: self._window.show_contact_detail(c)) if contact else None
            widget = MapView._build_marker_widget(r, g, b, icon_name, "", badge_text=badge)
            widget.set_tooltip_text(name)
            click = Gtk.GestureClick()
            click.connect(
                "pressed",
                lambda g,
                n,
                x,
                y,
                w=widget,
                nm=name,
                la=lat,
                lo=lon,
                t=type_label,
                ip=is_interpolated,
                od=detail_cb: MapView.show_marker_popover(
                    w, nm, la, lo, contact_type=t, interpolated=ip, on_detail=od
                ),
            )
            widget.add_controller(click)

            hop_marker = Shumate.Marker()
            hop_marker.set_child(widget)
            hop_marker.set_location(lat, lon)
            marker_layer.add_marker(hop_marker)

            lats.append(lat)
            lons.append(lon)

        MapView.fit_viewport(viewport, lats, lons)
        dialog.present(self._window)

    def _reply_to_message(self, msg):
        """Insert @[sender_name] into the text entry."""
        mention = f"@[{msg.sender_name}] " if msg.sender_name else ""
        self._text_entry.set_text(mention)
        self._text_entry.set_position(len(mention))
        self._text_entry.grab_focus()

    def _copy_message_text(self, msg):
        """Copy message text to clipboard."""
        clipboard = self._window.get_clipboard()
        clipboard.set(msg.text)

    def _show_repeats_dialog(self, key):
        repeats = self._repeats.get(key, [])
        if not repeats:
            dialog = Adw.AlertDialog(
                heading=_("Heard Repeats"),
                body=_("No repeats heard yet."),
            )
            dialog.add_response("ok", _("OK"))
            dialog.present(self._window)
            return

        dialog = Adw.AlertDialog(heading=_("Heard Repeats"))
        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=4)
        for i, rep in enumerate(repeats, 1):
            hops = rep.get("path_hops", [])
            if hops:
                hop_names = []
                for prefix in hops:
                    contact = self._window.find_contact_by_key_prefix(prefix.lower())
                    hop_names.append(contact.name if contact else prefix)
                path_str = " \u2192 ".join(hop_names)
            else:
                path_len = rep.get("path_len", 0)
                path_str = (
                    f'{path_len} hop{"s" if path_len != 1 else ""}' if path_len > 0 else "direct"
                )
            label = Gtk.Label(label=f"{i}. {path_str}", xalign=0, wrap=True)
            content.append(label)
        dialog.set_extra_child(content)
        dialog.add_response("ok", _("OK"))
        dialog.present(self._window)

    # ─── Message actions ─────────────────────────────────────

    def _resend_message(self, msg, item):
        if self._current_channel:
            new_ts = self._window.resend_channel_message(self._current_channel.index, msg.text)
            original_key = self._make_msg_key(msg.channel_index, msg.timestamp, msg.text)
            resend_key = (msg.channel_index, new_ts, msg.text)
            self._resend_ts_map[resend_key] = original_key
            item.props.resend_count = item.resend_count + 1

    def _delete_message_item(self, item):
        msg = item.message
        # Find and remove from store
        for i in range(self._store.get_n_items()):
            if self._store.get_item(i) is item:
                self._store.remove(i)
                break
        # Clean up tracking dicts
        key = self._make_msg_key(msg.channel_index, msg.timestamp, msg.text)
        self._outgoing_items.pop(key, None)
        self._incoming_items.pop(key, None)
        self._repeats.pop(key, None)
        self._incoming_paths.pop(key, None)
        self._seen_channel_msgs.discard(key)
        # Delete from storage
        self._window.delete_channel_message(
            msg.channel_index,
            msg.timestamp.timestamp(),
            msg.text,
        )

    # ─── Input ───────────────────────────────────────────────

    def _get_channel_limit(self) -> int:
        sender_name = self._window.device_info.name or ""
        return max_channel_text_bytes(sender_name)

    def _on_text_changed(self, entry):
        text = entry.get_text()
        used = len(text.encode("utf-8"))
        limit = self._get_channel_limit()
        if used == 0:
            self._char_label.set_label("")
            self._send_btn.set_sensitive(True)
            self._char_label.remove_css_class("error")
        else:
            self._char_label.set_label(f"{used}/{limit}")
            over = used > limit
            self._send_btn.set_sensitive(not over)
            if over:
                self._char_label.add_css_class("error")
            else:
                self._char_label.remove_css_class("error")
        self._check_mention_trigger(text, entry.get_position())

    def _on_send(self, *args):
        if not self._current_channel:
            return
        text = self._text_entry.get_text().strip()
        if not text:
            return
        if len(text.encode("utf-8")) > self._get_channel_limit():
            return
        self._text_entry.set_text("")
        self._window.send_channel_message(self._current_channel.index, text)

    # ─── External callbacks ──────────────────────────────────

    def on_channel_message_sent(self, msg: ChannelMessage, sender_ts: int):
        """Called when a channel message is sent by us."""
        if (
            self._current_channel
            and msg.channel_index == self._current_channel.index
            and self._stack.get_visible_child_name() == "chat"
        ):
            # Force sticky so notify::upper handler scrolls to the new message
            self._is_sticky = True
            self._add_message_item(msg)

    def _update_item_max_hops(self, key, path_info):
        item = self._incoming_items.get(key)
        if item and path_info:
            new_len = path_info.get("path_len", -1)
            if new_len > item.max_hops:
                item.props.max_hops = new_len

    def on_path_variant_received(self, msg, path_info=None):
        """Check if this is a duplicate message via a different path.
        Returns True if it's a duplicate (already seen), False if new."""
        key = self._make_msg_key(msg.channel_index, msg.timestamp, msg.text)
        if key in self._seen_channel_msgs:
            if path_info:
                self._incoming_paths.setdefault(key, []).append(path_info)
                self._update_item_max_hops(key, path_info)
            return True
        self._seen_channel_msgs.add(key)
        if path_info:
            self._incoming_paths.setdefault(key, []).append(path_info)
        return False

    def add_path_info(self, msg, path_info):
        """Append path_info to an existing message without affecting dedup."""
        key = self._make_msg_key(msg.channel_index, msg.timestamp, msg.text)
        if key in self._incoming_paths:
            self._incoming_paths[key].append(path_info)
        else:
            self._incoming_paths[key] = [path_info]
        self._update_item_max_hops(key, path_info)

    def on_channel_message_received(self, msg: ChannelMessage, path_info=None):
        if (
            self._current_channel
            and msg.channel_index == self._current_channel.index
            and self._stack.get_visible_child_name() == "chat"
        ):
            # If sticky, the notify::upper handler will auto-scroll
            self._add_message_item(msg)

    def on_repeat_received(self, channel_idx, sender_ts, text, path_hops):
        """Called when an echo/repeat of our outgoing message is received.

        Returns the original message key (channel_idx, ts, text) if matched,
        so the caller can persist path info under the original timestamp.
        """
        repeat_entry = {
            "path_hops": path_hops or [],
            "path_len": len(path_hops) if path_hops else 0,
        }

        # Check if this echo matches a resend timestamp
        resend_key = (channel_idx, sender_ts, text)
        matched_key = self._resend_ts_map.pop(resend_key, None)

        if matched_key is None:
            # Fall back to heuristic: same channel + timestamp within 30s + matching text
            for k in self._outgoing_items:
                if k[0] != channel_idx:
                    continue
                ts_diff = abs(k[1] - sender_ts)
                if ts_diff > 30:
                    continue
                if k[2] == text or text.startswith(k[2]) or k[2].startswith(text):
                    matched_key = k
                    break

        if matched_key is None:
            key = (channel_idx, sender_ts, text)
            if key not in self._repeats:
                self._repeats[key] = []
            self._repeats[key].append(repeat_entry)
            return None

        if matched_key not in self._repeats:
            self._repeats[matched_key] = []
        self._repeats[matched_key].append(repeat_entry)
        item = self._outgoing_items.get(matched_key)
        if item:
            item.props.repeat_count = len(self._repeats[matched_key])
        return matched_key


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/channels-view.ui")
class ChannelsView(Gtk.Box):
    """Channels list - shown in the sidebar."""

    __gtype_name__ = "MeshyChannelsView"

    _channel_list = Gtk.Template.Child("channel_list")
    _search_bar = Gtk.Template.Child("search_bar")
    _search_entry = Gtk.Template.Child("search_entry")
    _status_label = Gtk.Template.Child("status_label")

    FILTER_ALL = "all"
    FILTER_PUBLIC = "public"
    FILTER_HASHTAG = "hashtag"
    FILTER_PRIVATE = "private"

    SORT_AZ = "az"
    SORT_MESSAGES = "messages"

    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self._window = window
        self._chat_widget = ChannelChatWidget(window)
        self._unread_indices: set[int] = set()
        self._unread_counts: dict[int, int] = {}
        self._unread_dots: dict[int, Gtk.Label] = {}
        self._current_filter = self.FILTER_ALL
        settings = Gio.Settings(schema_id="page.codeberg.sesivany.Meshy")
        saved_sort = settings.get_string("channel-sort")
        self._current_sort = (
            saved_sort if saved_sort in (self.SORT_AZ, self.SORT_MESSAGES) else self.SORT_MESSAGES
        )
        self._total_channels = 0

        self._channel_list.connect("row-activated", self._on_row_selected)
        self._search_bar.connect_entry(self._search_entry)
        self._search_bar.connect("notify::search-mode-enabled", self._on_search_mode_changed)
        self._search_entry.connect("search-changed", self._on_search_changed)
        self._channel_list.set_filter_func(self._filter_func)

        # Actions for the add channel menu and filter/sort
        action_group = Gio.SimpleActionGroup()
        action = Gio.SimpleAction.new("create-private", None)
        action.connect("activate", self._on_create_private_channel)
        action_group.add_action(action)

        action = Gio.SimpleAction.new("join-private", None)
        action.connect("activate", self._on_join_private_channel)
        action_group.add_action(action)

        action = Gio.SimpleAction.new("join-public", None)
        action.connect("activate", self._on_join_public_channel)
        action_group.add_action(action)

        action = Gio.SimpleAction.new("join-hashtag", None)
        action.connect("activate", self._on_join_hashtag_channel)
        action_group.add_action(action)

        filter_action = Gio.SimpleAction.new_stateful(
            "filter", GLib.VariantType.new("s"), GLib.Variant.new_string(self._current_filter)
        )
        filter_action.connect("activate", self._on_filter_changed)
        action_group.add_action(filter_action)
        sort_action = Gio.SimpleAction.new_stateful(
            "sort", GLib.VariantType.new("s"), GLib.Variant.new_string(self._current_sort)
        )
        sort_action.connect("activate", self._on_sort_changed)
        action_group.add_action(sort_action)

        self._action_group = action_group
        self.insert_action_group("channel", action_group)
        self._search_btn = None
        self._search_binding = None

        self.update_channels(self._window.channels)

    def get_chat_widget(self) -> ChannelChatWidget:
        return self._chat_widget

    def set_search_button(self, btn):
        if self._search_binding:
            self._search_binding.unbind()
        self._search_btn = btn
        self._search_binding = btn.bind_property(
            "active",
            self._search_bar,
            "search-mode-enabled",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )

    def build_filter_menu(self):
        menu = Gio.Menu()
        for filter_id, label in (
            ("all", _("All")),
            ("public", _("Public")),
            ("hashtag", _("Hashtag")),
            ("private", _("Private")),
        ):
            item = Gio.MenuItem.new(label, None)
            item.set_action_and_target_value("channel.filter", GLib.Variant.new_string(filter_id))
            menu.append_item(item)
        return menu

    def build_sort_menu(self):
        menu = Gio.Menu()
        for sort_id, label in (
            ("az", _("A–Z")),
            ("messages", _("Latest Messages")),
        ):
            item = Gio.MenuItem.new(label, None)
            item.set_action_and_target_value("channel.sort", GLib.Variant.new_string(sort_id))
            menu.append_item(item)
        return menu

    def _filter_func(self, row) -> bool:
        if not hasattr(row, "_channel"):
            return True
        channel = row._channel
        f = self._current_filter
        if (
            f == self.FILTER_PUBLIC
            and not channel.is_public_channel
            or f == self.FILTER_HASHTAG
            and not channel.is_hashtag_channel
            or f == self.FILTER_PRIVATE
            and (channel.is_public_channel or channel.is_hashtag_channel)
        ):
            return False
        search = self._search_entry.get_text().lower()
        return not (search and search not in (channel.name or "").lower())

    def _update_status_count(self):
        total = self._total_channels
        if self._current_filter != self.FILTER_ALL or self._search_entry.get_text():
            visible = 0
            row = self._channel_list.get_first_child()
            while row:
                if hasattr(row, "_channel") and self._filter_func(row):
                    visible += 1
                row = row.get_next_sibling()
            self._status_label.set_label(
                _("{visible}/{total} channels").format(visible=visible, total=total)
            )
        else:
            self._status_label.set_label(
                _("{total} channels").format(total=total)
                if total != 1
                else _("{total} channel").format(total=total)
            )

    def _on_search_mode_changed(self, search_bar, pspec):
        if not search_bar.get_search_mode():
            self._search_entry.set_text("")

    def _on_search_changed(self, entry):
        self._channel_list.invalidate_filter()
        GLib.idle_add(self._update_status_count)

    def _on_filter_changed(self, action, param):
        self._current_filter = param.get_string()
        action.set_state(param)
        self._channel_list.invalidate_filter()
        GLib.idle_add(self._update_status_count)

    def _on_sort_changed(self, action, param):
        self._current_sort = param.get_string()
        action.set_state(param)
        settings = Gio.Settings(schema_id="page.codeberg.sesivany.Meshy")
        settings.set_string("channel-sort", self._current_sort)
        self.update_channels(self._window.channels)

    def _channel_sort_key(self, channel: Channel):
        type_order = 0 if channel.is_public_channel else (1 if channel.is_hashtag_channel else 2)
        if self._current_sort == self.SORT_AZ:
            return (type_order, (channel.name or "").lower())
        else:
            last_msg = self._chat_widget.get_last_message_time(channel.index)
            return (type_order, -(last_msg or 0))

    def _on_row_selected(self, listbox, row):
        if row and hasattr(row, "_channel"):
            self._chat_widget.set_channel(row._channel)
            # Navigate to content panel (matters in collapsed/phone mode)
            title = row._channel.name or _("Channel {}").format(row._channel.index)
            if row._channel.region:
                title = _("{name} ({region})").format(name=title, region=row._channel.region)
            self._window.content_page.set_title(title)
            self._window.split_view.set_show_content(True)
            self.mark_unread(row._channel.index, False)

    def update_channels(self, channels: list[Channel]):
        child = self._channel_list.get_first_child()
        while child:
            next_child = child.get_next_sibling()
            self._channel_list.remove(child)
            child = next_child

        non_empty = [c for c in channels if not c.is_empty]
        self._total_channels = len(non_empty)
        sorted_channels = sorted(non_empty, key=self._channel_sort_key)

        for channel in sorted_channels:
            row = Adw.ActionRow(
                title=channel.name or _("Channel {}").format(channel.index),
                subtitle=channel.region,
            )
            if channel.is_hashtag_channel:
                icon_name = "hashtag-symbolic"
            elif channel.is_public_channel:
                icon_name = "system-users-symbolic"
            else:
                icon_name = "padlock2-symbolic"
            row.add_prefix(Gtk.Image.new_from_icon_name(icon_name))
            row.set_activatable(True)
            row._channel = channel

            # Unread badge — visible by default when count > 0, hidden on hover
            count = self._unread_counts.get(channel.index, 0)
            unread_badge = Gtk.Label(label=str(count) if count else "")
            unread_badge.add_css_class("unread-badge")
            unread_badge.set_valign(Gtk.Align.CENTER)
            unread_badge.set_visible(count > 0)
            row.add_suffix(unread_badge)
            self._unread_dots[channel.index] = unread_badge

            # "More" button + hover behavior + long press
            info_btn = Gtk.Button(icon_name="view-more-symbolic", valign=Gtk.Align.CENTER)
            info_btn.add_css_class("flat")
            info_btn.set_visible(False)
            on_detail = lambda *_, ch=channel: self._show_channel_detail(ch)
            info_btn.connect("clicked", on_detail)
            row.add_suffix(info_btn)
            _add_hover_swap(
                row,
                info_btn,
                unread_badge,
                lambda _idx=channel.index: self._unread_counts.get(_idx, 0),
                on_detail,
            )

            self._channel_list.append(row)

        GLib.idle_add(self._update_status_count)

    def _get_visible_rows(self):
        """Return visible rows in display order."""
        rows = []
        row = self._channel_list.get_first_child()
        while row:
            if row.get_visible() and hasattr(row, "_channel"):
                rows.append(row)
            row = row.get_next_sibling()
        return rows

    def navigate(self, direction: int, unread_only: bool = False):
        """Move selection up (-1) or down (+1). If unread_only, skip to next unread."""
        rows = self._get_visible_rows()
        if not rows:
            return
        if unread_only:
            rows = [r for r in rows if r._channel.index in self._unread_indices]
            if not rows:
                return
        selected = self._channel_list.get_selected_row()
        if selected and selected in rows:
            idx = rows.index(selected)
            idx = (idx + direction) % len(rows)
        else:
            idx = 0 if direction > 0 else len(rows) - 1
        self._channel_list.select_row(rows[idx])
        rows[idx].grab_focus()

    def on_path_variant_received(self, msg, path_info=None):
        return self._chat_widget.on_path_variant_received(msg, path_info)

    def add_path_info(self, msg, path_info):
        self._chat_widget.add_path_info(msg, path_info)

    def on_channel_message_received(self, msg: ChannelMessage, path_info=None):
        self._chat_widget.on_channel_message_received(msg, path_info)

    def on_channel_message_sent(self, msg: ChannelMessage, sender_ts: int):
        self._chat_widget.on_channel_message_sent(msg, sender_ts)

    def on_repeat_received(self, channel_idx, sender_ts, text, path_hops):
        return self._chat_widget.on_repeat_received(channel_idx, sender_ts, text, path_hops)

    def mark_unread(self, channel_index: int, unread: bool):
        """Show or hide the unread badge for a channel."""
        if unread:
            self._unread_indices.add(channel_index)
            self._unread_counts[channel_index] = self._unread_counts.get(channel_index, 0) + 1
        else:
            self._unread_indices.discard(channel_index)
            self._unread_counts[channel_index] = 0
        if channel_index in self._unread_dots:
            badge = self._unread_dots[channel_index]
            count = self._unread_counts.get(channel_index, 0)
            badge.set_label(str(count) if count else "")
            badge.set_visible(count > 0)
        self._window.update_nav_unread_channels()

    def _show_channel_detail(self, channel: Channel):
        """Show channel detail dialog with rename, PSK, notifications, and remove."""
        dialog = Adw.Dialog()
        dialog.set_title(channel.name or _("Channel {}").format(channel.index))
        dialog.set_content_width(420)
        dialog.set_content_height(480)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        scroll = Gtk.ScrolledWindow(vexpand=True)
        clamp = Adw.Clamp(maximum_size=420)
        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=12,
            margin_start=12,
            margin_end=12,
            margin_top=12,
            margin_bottom=12,
        )

        # Info group
        info_group = Adw.PreferencesGroup(title=_("Channel Info"))

        if channel.is_public_channel:
            name_entry = Adw.EntryRow(title=_("Name"))
            name_entry.set_text(channel.name)
            name_entry.set_show_apply_button(True)

            def _on_rename_apply(*args):
                new_name = name_entry.get_text().strip()
                if new_name:
                    self._window.set_channel(channel.index, new_name, channel.psk)

            name_entry.connect("apply", _on_rename_apply)
            info_group.add(name_entry)
        else:
            info_group.add(
                Adw.ActionRow(
                    title=_("Name"),
                    subtitle=channel.name or _("Channel {}").format(channel.index),
                )
            )

        info_group.add(
            Adw.ActionRow(
                title=_("Type"),
                subtitle=channel.channel_type,
            )
        )

        # Private Key row with copy button (only for private channels)
        if not channel.is_public_channel and not channel.is_hashtag_channel:
            psk_row = Adw.ActionRow(
                title=_("Private Key"),
                subtitle=channel.psk_hex,
            )
            psk_row.set_subtitle_selectable(True)
            copy_btn = Gtk.Button(icon_name="edit-copy-symbolic", valign=Gtk.Align.CENTER)
            copy_btn.add_css_class("flat")
            copy_btn.set_tooltip_text("Copy private key to clipboard")

            def _on_copy_psk(*args):
                clipboard = self._window.get_clipboard()
                clipboard.set(channel.psk_hex)
                self._window.show_toast(_("Private key copied to clipboard"), timeout=2)

            copy_btn.connect("clicked", _on_copy_psk)
            psk_row.add_suffix(copy_btn)
            info_group.add(psk_row)

        content.append(info_group)

        # Notifications group
        notif_group = Adw.PreferencesGroup(title=_("Notifications"))

        notif_row = Adw.ComboRow(title=_("Notification Level"))
        notif_list = Gtk.StringList()
        for label in [_("Default"), _("All Messages"), _("Mentions Only"), _("None")]:
            notif_list.append(label)
        notif_row.set_model(notif_list)

        level_to_index = {
            NotificationLevel.DEFAULT: 0,
            NotificationLevel.ALL: 1,
            NotificationLevel.MENTIONS: 2,
            NotificationLevel.NONE: 3,
        }
        notif_row.set_selected(level_to_index.get(channel.notification_level, 0))

        def _on_notif_changed(*args):
            selected = notif_row.get_selected()
            level = [
                NotificationLevel.DEFAULT,
                NotificationLevel.ALL,
                NotificationLevel.MENTIONS,
                NotificationLevel.NONE,
            ][selected]
            self._window.set_channel_notification_level(channel, level)

        notif_row.connect("notify::selected", _on_notif_changed)
        notif_group.add(notif_row)

        content.append(notif_group)

        # Region group
        available_regions = self._window.get_regions()
        region_group = Adw.PreferencesGroup(
            title=_("Region"),
            description=_("Limit flood messages to a specific region"),
        )

        if available_regions:
            region_row = Adw.ComboRow(title=_("Region"))
            region_list = Gtk.StringList()
            region_list.append(_("None"))
            for r in available_regions:
                region_list.append(r)
            region_row.set_model(region_list)

            if channel.region in available_regions:
                region_row.set_selected(available_regions.index(channel.region) + 1)
            else:
                region_row.set_selected(0)

            def _on_region_changed(*args):
                selected = region_row.get_selected()
                new_region = available_regions[selected - 1] if selected > 0 else ""
                channel.region = new_region
                if self._window.storage:
                    self._window.storage.set_channel_region(channel.index, new_region)
                self.update_channels(self._window._channels)

            region_row.connect("notify::selected", _on_region_changed)
            region_group.add(region_row)
        else:
            hint_row = Adw.ActionRow(
                title=_("No regions defined"),
                subtitle=_("Add regions in Settings to assign them to channels"),
            )
            hint_row.add_suffix(Gtk.Image(icon_name="go-next-symbolic"))
            hint_row.set_activatable(True)

            def _on_hint_activated(*args):
                dialog.connect(
                    "closed", lambda *a: self._window._nav_buttons["settings"].set_active(True)
                )
                dialog.close()

            hint_row.connect("activated", _on_hint_activated)
            region_group.add(hint_row)

        content.append(region_group)

        # Actions group
        actions_group = Adw.PreferencesGroup(title=_("Actions"))

        remove_row = Adw.ActionRow(title=_("Remove Channel"))
        remove_btn = Gtk.Button(icon_name="user-trash-symbolic", valign=Gtk.Align.CENTER)
        remove_btn.add_css_class("destructive-action")

        def _on_remove(*args):
            confirm = Adw.AlertDialog(
                heading=_("Remove Channel?"),
                body=_('Remove "{}" and all its messages?').format(
                    channel.name or _("Channel {}").format(channel.index)
                ),
            )
            confirm.add_response("cancel", _("Cancel"))
            confirm.add_response("remove", _("Remove"))
            confirm.set_response_appearance("remove", Adw.ResponseAppearance.DESTRUCTIVE)
            confirm.connect(
                "response",
                lambda d, r: (self._window.remove_channel(channel), dialog.close())
                if r == "remove"
                else None,
            )
            confirm.present(self._window)

        remove_btn.connect("clicked", _on_remove)
        remove_row.add_suffix(remove_btn)
        remove_row.set_activatable_widget(remove_btn)
        actions_group.add(remove_row)

        content.append(actions_group)

        clamp.set_child(content)
        scroll.set_child(clamp)
        toolbar_view.set_content(scroll)
        dialog.set_child(toolbar_view)
        dialog.present(self._window)

    def _find_next_empty_index(self) -> int | None:
        """Find the next available channel index (0-39)."""
        used = {ch.index for ch in self._window.channels if not ch.is_empty}
        for i in range(40):
            if i not in used:
                return i
        return None

    def _show_no_slots_dialog(self):
        dialog = Adw.AlertDialog(
            heading=_("No Slots Available"),
            body=_("All 8 channel slots are in use."),
        )
        dialog.add_response("ok", _("OK"))
        dialog.present(self._window)

    def _on_create_private_channel(self, *args):
        idx = self._find_next_empty_index()
        if idx is None:
            self._show_no_slots_dialog()
            return

        dialog = Adw.AlertDialog(
            heading=_("Create a Private Channel"),
            body=_("A random PSK will be generated. Share it with others so they can join."),
        )
        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)
        name_row = Adw.EntryRow(title=_("Channel Name"))
        fields_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        fields_list.add_css_class("boxed-list")
        fields_list.append(name_row)
        content.append(fields_list)
        dialog.set_extra_child(content)
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("create", _("Create"))
        dialog.set_response_appearance("create", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response == "create":
                import os

                name = name_row.get_text().strip()
                if not name:
                    return
                psk = os.urandom(16)
                self._window.set_channel(idx, name, psk)

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _on_join_private_channel(self, *args):
        idx = self._find_next_empty_index()
        if idx is None:
            self._show_no_slots_dialog()
            return

        dialog = Adw.AlertDialog(
            heading=_("Join a Private Channel"),
            body=_("Enter the channel name and private key shared with you."),
        )
        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)
        name_row = Adw.EntryRow(title=_("Channel Name"))
        psk_row = Adw.EntryRow(title=_("Private Key (32 hex characters)"))
        fields_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        fields_list.add_css_class("boxed-list")
        fields_list.append(name_row)
        fields_list.append(psk_row)
        content.append(fields_list)
        dialog.set_extra_child(content)
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("join", _("Join"))
        dialog.set_response_appearance("join", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response == "join":
                try:
                    name = name_row.get_text().strip()
                    psk = bytes.fromhex(psk_row.get_text().strip())
                    if len(psk) == 16 and name:
                        self._window.set_channel(idx, name, psk)
                except ValueError:
                    pass

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _on_join_public_channel(self, *args):
        # Check if already on the public channel
        for ch in self._window.channels:
            if ch.is_public_channel:
                dialog = Adw.AlertDialog(
                    heading=_("Already Joined"), body=_("You are already on the public channel.")
                )
                dialog.add_response("ok", _("OK"))
                dialog.present(self._window)
                return

        idx = self._find_next_empty_index()
        if idx is None:
            self._show_no_slots_dialog()
            return

        psk = bytes.fromhex(Channel.PUBLIC_CHANNEL_PSK)
        self._window.set_channel(idx, _("Public"), psk)

    def _on_join_hashtag_channel(self, *args):
        idx = self._find_next_empty_index()
        if idx is None:
            self._show_no_slots_dialog()
            return

        dialog = Adw.AlertDialog(
            heading=_("Join a Hashtag Channel"),
            body=_("Enter a hashtag name. Anyone using the same hashtag can communicate."),
        )
        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)
        hashtag_row = Adw.EntryRow(title=_("Hashtag (e.g. #general)"))
        fields_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        fields_list.add_css_class("boxed-list")
        fields_list.append(hashtag_row)
        content.append(fields_list)
        dialog.set_extra_child(content)
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("join", _("Join"))
        dialog.set_response_appearance("join", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response == "join":
                tag = hashtag_row.get_text().strip()
                if not tag:
                    return
                if not tag.startswith("#"):
                    tag = f"#{tag}"
                h = hashlib.sha256(tag.encode()).digest()
                psk = h[:16]
                name = tag
                self._window.set_channel(idx, name, psk)

        dialog.connect("response", _on_response)
        dialog.present(self._window)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import sys

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")

from meshy.application import MeshyApplication


def main(version):
    app = MeshyApplication(version=version)
    return app.run(sys.argv)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Path history tracking and smart route selection for private messages."""

from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from meshy.storage import Storage

log = logging.getLogger(__name__)

WEIGHT_SUCCESS_INCREMENT = 0.5
WEIGHT_FAILURE_DECREMENT = 0.2
MAX_WEIGHT = 5.0


@dataclass
class PathRecord:
    id: int
    contact_key_hex: str
    path_bytes: str
    path_hash_size: int
    hop_count: int
    success_count: int
    failure_count: int
    last_trip_ms: int | None
    route_weight: float
    last_success_at: float | None


class PathHistoryService:
    def __init__(self, storage: Storage):
        self._storage = storage

    def record_result(
        self,
        contact_key_hex: str,
        path_bytes: str,
        path_hash_size: int,
        hop_count: int,
        success: bool,
        trip_ms: int | None = None,
        weight_increment: float | None = None,
    ):
        if weight_increment is None:
            weight_increment = WEIGHT_SUCCESS_INCREMENT if success else -WEIGHT_FAILURE_DECREMENT
        self._storage.upsert_path_record(
            contact_key_hex,
            path_bytes,
            path_hash_size,
            hop_count,
            success,
            trip_ms,
            weight_increment=weight_increment,
            max_weight=MAX_WEIGHT,
        )
        log.debug(
            "Path %s for %s: %s (trip=%s)",
            path_bytes,
            contact_key_hex[:8],
            "success" if success else "failure",
            trip_ms,
        )

    FIRMWARE_DISCOVERY_WEIGHT = 1.0

    def seed_from_contact(
        self,
        contact_key_hex: str,
        path_bytes: str,
        path_hash_size: int,
        hop_count: int,
    ):
        existing = [
            r
            for r in self._storage.get_path_history(contact_key_hex)
            if r["path_bytes"] == path_bytes
        ]
        if existing:
            return
        self._storage.upsert_path_record(
            contact_key_hex,
            path_bytes,
            path_hash_size,
            hop_count,
            success=True,
            initial_weight=self.FIRMWARE_DISCOVERY_WEIGHT,
        )
        log.debug(
            "Seeded path %s for %s from firmware (weight=%.1f)",
            path_bytes,
            contact_key_hex[:8],
            self.FIRMWARE_DISCOVERY_WEIGHT,
        )

    def has_paths(self, contact_key_hex: str) -> bool:
        return bool(self._storage.get_path_history(contact_key_hex))

    def select_path(
        self,
        contact_key_hex: str,
        exclude_paths: list[str] | None = None,
    ) -> PathRecord | None:
        records = [self._dict_to_record(r) for r in self._storage.get_path_history(contact_key_hex)]
        if not records:
            return None

        if exclude_paths:
            candidates = [r for r in records if r.path_bytes not in exclude_paths]
            if not candidates:
                return None
        else:
            candidates = records

        return max(candidates, key=self.score, default=None)

    def score(self, record: PathRecord) -> float:
        total = record.success_count + record.failure_count
        reliability = (record.success_count + 1) / (total + 2)

        if record.last_trip_ms and record.last_trip_ms > 0:
            latency = 1.0 / (1.0 + record.last_trip_ms / 10000.0)
        else:
            latency = 0.5

        if record.last_success_at:
            age_days = (time.time() - record.last_success_at) / 86400.0
            freshness = 1.0 / (1.0 + age_days)
        else:
            freshness = 0.0

        return reliability * 0.6 + latency * 0.2 + freshness * 0.2

    def score_dict(self, d: dict) -> float:
        return self.score(self._dict_to_record(d))

    @staticmethod
    def _dict_to_record(d: dict) -> PathRecord:
        return PathRecord(
            id=d["id"],
            contact_key_hex=d["contact_key_hex"],
            path_bytes=d["path_bytes"],
            path_hash_size=d["path_hash_size"],
            hop_count=d["hop_count"],
            success_count=d["success_count"],
            failure_count=d["failure_count"],
            last_trip_ms=d["last_trip_ms"],
            route_weight=d["route_weight"],
            last_success_at=d["last_success_at"],
        )



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shared behaviour for chat-style ListViews (DM chat and channel chat)."""

from gi.repository import GLib, Gtk


class ChatScrollMixin:
    """Scroll management, context-menu helpers, and emoji picker.

    Host class must provide:
        _scroll             Gtk.ScrolledWindow
        _list_view          Gtk.ListView
        _store              Gio.ListStore
        _scroll_down_revealer  Gtk.Revealer
        _unread_btn_revealer   Gtk.Revealer
        _text_entry         Gtk.Entry / Gtk.Text
        _context_popover    Gtk.PopoverMenu
        _window             main application window
        _is_sticky          bool
        _restoring_scroll   bool
        _context_menu_open  bool
        _frozen_scroll_value float
        _unread_pos         int
        _load_older_messages()  method
    """

    # ─── Scroll state ────────────────────────────────────────

    def _is_at_bottom(self) -> bool:
        adj = self._scroll.get_vadjustment()
        return adj.get_value() + adj.get_page_size() >= adj.get_upper() - 1.0

    def _on_edge_overshot(self, scrolled_window, pos):
        if pos == Gtk.PositionType.TOP:
            self._load_older_messages()

    def _on_scroll_value_changed(self, adj):
        if self._restoring_scroll:
            return
        if self._context_menu_open:
            adj.set_value(self._frozen_scroll_value)
            return
        at_bottom = self._is_at_bottom()
        self._is_sticky = at_bottom
        self._scroll_down_revealer.set_reveal_child(not at_bottom)
        self._update_unread_btn_visibility()

    def _on_scroll_upper_changed(self, adj, *args):
        if self._restoring_scroll:
            return
        if self._context_menu_open:
            adj.set_value(self._frozen_scroll_value)
            return
        if self._is_sticky:
            self._scroll_down()
        else:
            self._scroll_down_revealer.set_reveal_child(not self._is_at_bottom())

    def _is_item_visible(self, position: int) -> bool:
        adj = self._scroll.get_vadjustment()
        top = adj.get_value()
        bottom = top + adj.get_page_size()
        n = self._store.get_n_items()
        if position < 0 or position >= n or n == 0:
            return False
        item_y = (position / n) * adj.get_upper()
        return top <= item_y < bottom

    def _update_unread_btn_visibility(self):
        if self._unread_pos < 0:
            self._unread_btn_revealer.set_reveal_child(False)
            return
        visible = self._is_item_visible(self._unread_pos)
        if visible:
            self._unread_pos = -1
            self._unread_btn_revealer.set_reveal_child(False)
        else:
            self._unread_btn_revealer.set_reveal_child(True)

    # ─── Scroll actions ──────────────────────────────────────

    def _scroll_down(self):
        if self._context_menu_open or self._is_at_bottom():
            return
        n = self._store.get_n_items()
        if n > 0:
            GLib.idle_add(self._do_scroll_down)

    def _do_scroll_down(self):
        if self._context_menu_open:
            return False
        n = self._store.get_n_items()
        if n > 0:
            self._list_view.scroll_to(n - 1, Gtk.ListScrollFlags.FOCUS, None)
        return False

    def _scroll_to_bottom(self):
        adj = self._scroll.get_vadjustment()
        adj.set_value(adj.get_upper())
        return False

    def _restore_scroll_position(self, value):
        adj = self._scroll.get_vadjustment()
        upper = adj.get_upper() - adj.get_page_size()
        adj.set_value(min(value, upper))
        self._restoring_scroll = False
        self._is_sticky = self._is_at_bottom()
        self._scroll_down_revealer.set_reveal_child(not self._is_sticky)
        return False

    def _scroll_to_item(self, position):
        n = self._store.get_n_items()
        if position < 0 or position >= n:
            return False
        self._list_view.scroll_to(position, Gtk.ListScrollFlags.FOCUS, None)
        return False

    def _on_scroll_to_bottom_clicked(self, btn):
        self._is_sticky = True
        self._scroll_down()

    def _on_scroll_to_unread_clicked(self, btn):
        if self._unread_pos >= 0:
            self._scroll_to_item(self._unread_pos)
            self._unread_pos = -1
            self._unread_btn_revealer.set_reveal_child(False)

    # ─── Context menu helpers ─────────────────────────────────

    def _find_outer_at(self, x, y):
        target = self._list_view.pick(x, y, Gtk.PickFlags.DEFAULT)
        while target is not None and target is not self._list_view:
            if hasattr(target, "_chat_item"):
                return target
            target = target.get_parent()
        return None

    def _on_context_popover_closed(self, popover):
        self._context_menu_open = False
        self._is_sticky = self._is_at_bottom()

    def _on_list_long_press(self, gesture, x, y):
        self._on_list_right_click(None, 1, x, y)

    # ─── Emoji ───────────────────────────────────────────────

    def _on_emoji_clicked(self, btn):
        chooser = Gtk.EmojiChooser()
        chooser.set_parent(btn)
        chooser.connect(
            "emoji-picked",
            lambda c, emoji: self._text_entry.insert_text(emoji, self._text_entry.get_position()),
        )
        chooser.connect("closed", lambda c: c.unparent())
        chooser.popup()



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Platform-agnostic camera/QR scanner service."""

from meshy.platforms import is_macos


def create_qr_scanner(window, on_result):
    """Create platform-appropriate QR scanner dialog.

    Args:
        window: Parent GTK window
        on_result: Callback function(data: str) called with QR code data

    Returns:
        QR scanner dialog instance with start() method
    """
    if is_macos():
        from meshy.platforms.macos.camera_avfoundation import QRScannerDialogMacOS
        return QRScannerDialogMacOS(window, on_result)
    else:
        from meshy.qr_scanner import QRScannerDialog
        return QRScannerDialog(window, on_result)
